import numpy as np
from pathlib import Path # For efficient path manipulation
import papylio as pp
import matplotlib.pyplot as plt
import tqdm
# from git import Repo
import time
import xarray as xr
from skimage.transform import AffineTransform

from papylio.plugins.sequencing.fastqAnalysis import FastqData
import matchpoint as mp

# This part makes an analysis.txt file in the same directory as the this .py file. (please do not commit this)
# This file contains the git commit used and all differences to this commit.
# writePath = Path(Path(__file__).parent)
# writeFile = writePath.joinpath('git_repository_version_and_differences.txt')
# repo = Repo(Path(pp.__file__).parent.parent)
#
# with writeFile.open("a") as f:
#     f.write('------------------------------------------\n\n')
#     f.write(f'Trace_analysis version: {repo.head.object.hexsha} \n\n')
#     t = repo.head.commit.tree
#     f.write(repo.git.diff(t))
#     f.write('\n\n------------------------------------------\n\n')



#####################################
# SINGLE-MOLECULE DATA PROCESSING
#####################################

experiment_path = r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20221023 - Objective-type TIRF (BN)'
# experiment_path = r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20221023 - Objective-type TIRF (BN)\Test'
exp = pp.Experiment(experiment_path)

files_channel_mapping = exp.files[exp.files.relativeFilePath.str.regex('Bead slide')]
files_green_laser = exp.files[exp.files.relativeFilePath.str.regex('3.*HJ_scan TIRF 561')]
files_red_laser_after = exp.files[exp.files.relativeFilePath.str.regex('3.*HJ_scan TIRF 642 after')]
files_red_laser_before = exp.files[exp.files.relativeFilePath.str.regex('3.*HJ_scan TIRF 642 before')]

files_red_laser_before.movie.illumination_arrangement = [1]
files_red_laser_after.movie.illumination_arrangement = [1]

# -----------------------------------
# Channel mapping
# -----------------------------------
# channel_mapping_file = files_channel_mapping[0]
# channel_mapping_file.perform_mapping()
#
# channel_mapping_file.show_average_image()
# channel_mapping_file.mapping.show_mapping_transformation(figure=plt.gcf(), show_source=True)

# -----------------------------------
# Image corrections
# -----------------------------------
exp.determine_flatfield_and_darkfield_corrections(files_green_laser, illumination_index=0, l_s=5, l_d=5)
exp.determine_flatfield_and_darkfield_corrections(files_red_laser_before, illumination_index=1, l_s=5, l_d=5)

files_green_laser.movie.determine_temporal_background_correction('fit_background_peak')
files_red_laser_before.movie.determine_temporal_background_correction('fit_background_peak')
files_red_laser_after.movie.determine_temporal_background_correction('fit_background_peak')

for file in tqdm.tqdm(files_green_laser):
    temporal_illumination_correction = file.movie.temporal_background_correction.sel(channel=1) / 1500
    temporal_background_correction = file.movie.temporal_background_correction / temporal_illumination_correction
    file.movie.temporal_background_correction = temporal_background_correction
    file.movie.temporal_illumination_correction = temporal_illumination_correction
    file.movie.save_corrections('temporal_background_correction', 'temporal_illumination_correction')

# -----------------------------------
# Find coordinates, extract traces and determine kinetics
# -----------------------------------

files_green_laser.find_coordinates()
files_green_laser[652].show_coordinates_in_image()

files_green_laser.extract_traces()


def add_coordinates_to_files(file_green, file_red_before, file_red_after):
    if (file_green.name[-4:] == file_red_before.name[-4:] == file_red_after.name[-4:]) and \
            (file_green.relativePath.name == file_red_before.relativePath.name == file_red_after.relativePath.name):
        file_red_before.coordinates = file_green.coordinates
        file_red_after.coordinates = file_green.coordinates
    else:
        raise FileNotFoundError

# TODO: Make this internal somehow
import joblib
from objectlist.base import tqdm_joblib
# for file_green, file_red_before, file_red_after in
with tqdm_joblib(tqdm.tqdm(total=len(files_green_laser))):
    joblib.parallel.Parallel(6)(joblib.parallel.delayed(add_coordinates_to_files)(*datum) for datum in
                                zip(files_green_laser, files_red_laser_before, files_red_laser_after))

files_red_laser_before.extract_traces()
files_red_laser_after.extract_traces()

def add_red_laser_intensity_to_files_green(file_green, file_red_before, file_red_after):
    if (file_green.name[-4:] == file_red_before.name[-4:] == file_red_after.name[-4:]) and \
            (file_green.relativePath.name == file_red_before.relativePath.name == file_red_after.relativePath.name):
        intensity_red_before = file_red_before.intensity.mean('frame')
        intensity_red_before = intensity_red_before.drop('file')
        intensity_red_before.name = 'intensity_red_before'
        intensity_red_before.to_netcdf(file_green.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')

        intensity_red_after = file_red_after.intensity.mean('frame')
        intensity_red_after = intensity_red_after.drop('file')
        intensity_red_after.name = 'intensity_red_after'
        intensity_red_after.to_netcdf(file_green.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')

    else:
        raise FileNotFoundError


with tqdm_joblib(tqdm.tqdm(total=len(files_green_laser))):
    joblib.parallel.Parallel(6)(joblib.parallel.delayed(add_red_laser_intensity_to_files_green)(*datum) for datum in
                                zip(files_green_laser, files_red_laser_before, files_red_laser_after))




#####################################
# IMPORT SEQUENCING DATA
#####################################

sequencing_analysis_path = Path(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20221024 - Sequencer (MiSeq)\Analysis')
aligned_sam_filepath = sequencing_analysis_path.joinpath('Alignment_read1.sam')
index1_fastq_filepath = sequencing_analysis_path.joinpath('Index1.fastq') # Can be None
extract_sequence_subset = [30, 31, 56, 57, 82, 83, 108, 109]

exp.import_sequencing_data(aligned_sam_filepath, index1_file_path=index1_fastq_filepath, remove_duplicates=True,
                           add_aligned_sequence=True, extract_sequence_subset=extract_sequence_subset)

#####################################
# SEQUENCING AND SINGLE-MOLECULE MAPPING
#####################################

mapping_sequence_name = 'HJ_general'

# files_scan123 = files_red_laser_before[files_red_laser_before.relativePath.str.regex('Scan 1|Scan 3')]
exp.generate_tile_mappings(files_green_laser, mapping_sequence_name=mapping_sequence_name, surface=0)

# files_scan4 = files_red_laser_before[files_red_laser_before.relativePath.str.regex('Scan 4')]
# exp.generate_tile_mappings(files_scan4, mapping_sequence_name=mapping_sequence_name, surface=0, name='Scan 4')


# -----------------------------------
# Finding rotation and scale with geometric hashing
# -----------------------------------

# TODO: Geometric hashing example here
# TODO: Put geometric hashing in MatchPoint



# -----------------------------------
# Finding translation with cross correlation
# -----------------------------------

# Previously found transformation
# Transformation based on sm pixel to sequencing MiSeq mapping with
# 'rotation': 0.006500218506032994, 'scale': [3.697153992993506, -3.697153992993506] using pixel size 0.125 µm
# exp.tile_mappings.transformation = AffineTransform(scale=[29.57723194394805, -29.57723194394805], rotation=0.006500218506032994)
exp.tile_mappings.transformation = AffineTransform(scale=[29.51, -29.51], rotation=0.0024)

exp.tile_mappings.use_parallel_processing = False
exp.tile_mappings.cross_correlation(peak_detection='auto', gaussian_width=7, divider=20, plot=False)

# bounds = ((0.98, 1.2), (-0.005, 0.005), (-250, 250), (-250, 250))
# exp.tile_mappings[19:(19+12)].kernel_correlation(bounds, sigma=25, crop='source',
#                                      strategy='best1bin', maxiter=1000, popsize=50, tol=0.01,
#                                      mutation=0.25, recombination=0.7, seed=None, callback=None, disp=False,
#                                      polish=False, init='sobol', atol=0, updating='immediate', workers=1,
#                                      constraints=())

exp.tile_mappings.save()
exp.tile_mappings.show_mapping_transformation(crop='source', save=True)

# -----------------------------------
# Determine translations for all tiles
# -----------------------------------
exp.tile_mappings.scatter_parameters('translation', 'translation', 'x', 'y', save=True)  # To determine correct mapping indices
exp.tile_mappings.estimate_translations(indices=np.arange(11), save=True)
exp.tile_mappings.scatter_parameters('translation', 'translation', 'x', 'y', save=True) # Rename original file before saving
exp.tile_mappings.save()
exp.tile_mappings.show_mapping_transformation(crop='source', save=True)

# -----------------------------------
# Matching statistics
# -----------------------------------

# TODO: Show how to extract matching statics, i.e. numbers/percentages of points matched.

# -----------------------------------
# Obtain file sequencing data and generate sequencing match
# -----------------------------------
files = files_green_laser[0:14*30]

# TODO: Make this automatically save the sequencing data
# TODO: Automatic import and export when setting or getting sequencing data
files_green_laser.parallel_processing_kwargs['require'] = 'sharedmem'
# files_green_scan123 = files_green_laser[files_green_laser.relativePath.str.regex('Scan 1|Scan 3')]
files_green_laser.get_sequencing_data(margin=5, mapping_name='All files')
# files_green_scan4 = files_green_laser[files_green_laser.relativePath.str.regex('Scan 4')]
# files_green_scan4.get_sequencing_data(margin=5, mapping_name='Scan 4 - HJ general')

files_green_laser.parallel_processing_kwargs.pop('require')
files_green_laser.generate_sequencing_match(overlapping_points_threshold=25,
                                            excluded_sequence_names=['MapSeq', 'CalSeq', '*'], plot=False)

# -----------------------------------
# Finetune the sequencing matches
# -----------------------------------

sequencing_matches = exp.sequencing_matches(files_green_laser)

sequencing_matches.cross_correlation(divider=1/10, gaussian_width=7, crop=True, plot=False)


sequencing_matches.transformation = AffineTransform()
sequencing_matches.transformation_inverse = AffineTransform()
bounds = ((0.97, 1.03), (-0.05, 0.05), (-5, 5), (-5, 5))
sequencing_matches.kernel_correlation(bounds, sigma=0.125, crop=True,
                                         strategy='best1bin', maxiter=1000, popsize=50, tol=0.01,
                                         mutation=0.25, recombination=0.7, seed=None, callback=None, disp=False,
                                         polish=True, init='sobol', atol=0, updating='immediate', workers=1,
                                         constraints=())

bounds = ((0.99, 1.01), (-0.01, 0.01), (-1, 1), (-1, 1))
sequencing_matches.kernel_correlation(bounds, sigma=0.125, crop=True,
                                         strategy='best1bin', maxiter=1000, popsize=50, tol=0.001,
                                         mutation=0.25, recombination=0.7, seed=None, callback=None, disp=False,
                                         polish=True, init='sobol', atol=0, updating='immediate', workers=1,
                                         constraints=())
sequencing_matches.save()

# -----------------------------------
# Find pairs and insert sequencing data into file dataset
# -----------------------------------

sequencing_matches.find_distance_threshold(maximum_radius=2)
sequencing_matches.determine_matched_pairs()
plt.figure()
plt.hist(np.hstack(sequencing_matches.pair_distances()), bins=100)
plt.title('Pair distances for istance_threshold 0.42 um')
plt.xlabel('Distance (um)')
plt.ylabel('Count')

sequencing_matches.destination_distance_threshold = 0.2  # 0.506
sequencing_matches.determine_matched_pairs()
sequencing_matches.save()

sequencing_matches.show_mapping_transformation()

files_green_laser.insert_sequencing_data_into_file_dataset(include_raw_sequences=False, include_aligned_sequences=True,
                                                           include_sequence_subset=True, determine_matched_pairs=True)


source_cropped_lengths = sequencing_matches.source_cropped.map(len)()
destination_cropped_lengths = sequencing_matches.destination_cropped.map(len)()
matched_pairs_lenghts = sequencing_matches.number_of_matched_points


source_transformed_combined = np.vstack([sequencing_match.transformation(sequencing_match.source) for sequencing_match in sequencing_matches])
destination_combined = np.unique(np.vstack([sequencing_match.destination for sequencing_match in sequencing_matches]), axis=0)
test = mp.MatchPoint(source=source_transformed_combined, destination=destination_combined)
test.destination_distance_threshold = 0.2
test.determine_pairs()

# ds = xr.open_mfdataset([file.relativeFilePath.with_suffix('.nc') for file in files_green_laser[0:300]], combine='nested', concat_dim='molecule',
#                        data_vars='minimal', coords='minimal', compat='override', engine='h5netcdf')#, parallel=True)#, chunks={'molecule': 100})
import time
start = time.time()
ds = xr.open_mfdataset([file.relativeFilePath.with_suffix('.nc') for file in files_green_laser], combine='nested', concat_dim='molecule',
                       data_vars='minimal', coords='minimal', compat='override', engine='h5netcdf', parallel=False)
print(time.time()-start)


# dataset_complete_subset = ds.sel(molecule=~ds.sequence_subset.str.contains(b'-'))
#
# folder = exp.main_path.joinpath('Analysis').joinpath('complete_dataset.nc')
# for group, value in tqdm.tqdm(ds.groupby('sequence_subset')):
#     if b'-' not in group and group != b'':
#         value.to_netcdf(folder.joinpath(group.decode()).with_suffix('.nc'), engine='h5netcdf')





import time
start = time.time()
ds.to_netcdf(exp.main_path.joinpath('Analysis').joinpath('complete_dataset.nc'), engine='h5netcdf', mode='w')
print(time.time()-start)



stage_coordinates = files_green_laser[(38*30):((38+15)*30)].movie.stage_coordinates

tile = exp.sequencing_data.sel(tile=1104)
ch = scipy.spatial.ConvexHull(tile.coordinates, incremental=False, qhull_options=None)


hull_coordinates = tile.coordinates[ch.vertices]
hull_coordinates = np.vstack([hull_coordinates.values, hull_coordinates.values[0]])
hull_coordinates_um = exp.tile_mappings[np.where((np.array(exp.tile_mappings.tile)==1101))[0].item()].transform_coordinates(hull_coordinates, inverse=True)

plt.figure()
hull_coordinates_um[:, 1] += 3200
hull_coordinates_um[:, 0] += 40
plt.plot(*hull_coordinates_um.T, c='r')
outline = np.array([[0, 0], [0, 64], [32, 64], [32, 0], [0, 0]])
stage_coordinates_adapted = np.flip(np.vstack(stage_coordinates), axis=1)
stage_coordinates_adapted[:,1] += 3200
for sc in stage_coordinates_adapted:
    plt.plot(*(sc + outline).T, c='g')
plt.gca().set_aspect('equal', 'box')
plt.xlabel('x (μm)')
plt.ylabel('y (μm)')
plt.tight_layout()



axis('equal')


ds = xr.open_dataset(exp.main_path.joinpath('Analysis').joinpath('complete_dataset.nc'), engine='h5netcdf')#, chunks={'molecule': 2000})




sequence_subset_hj1 = b'TTAGCCGA' # HJ1
sequence_subset_hj3 = b'AATCGGCT' # HJ3
sequence_subset_hj7 = b'GGCGCCGC' # HJ7






sequence_mutations_hj7 = ['GGCGCCGC', 'AGCGCCGC', 'CGCGCCGC', 'TGCGCCGC', 'AACGCCGC', 'ACCGCCGC', 'ATCGCCGC', 'CACGCCGC', 'CCCGCCGC', 'CTCGCCGC', 'TACGCCGC', 'TCCGCCGC', 'TTCGCCGC', 'AGAGCCGC', 'AGTGCCGC', 'AGGGCCGC', 'CGAGCCGC', 'CGTGCCGC', 'CGGGCCGC', 'TGAGCCGC', 'TGTGCCGC', 'TGGGCCGC', 'AGCACCGC', 'AGCCCCGC', 'AGCTCCGC', 'CGCACCGC', 'CGCCCCGC', 'CGCTCCGC', 'TGCACCGC', 'TGCCCCGC', 'TGCTCCGC', 'AGCGACGC', 'AGCGTCGC', 'AGCGGCGC', 'CGCGACGC', 'CGCGTCGC', 'CGCGGCGC', 'TGCGACGC', 'TGCGTCGC', 'TGCGGCGC', 'AGCGCAGC', 'AGCGCTGC', 'AGCGCGGC', 'CGCGCAGC', 'CGCGCTGC', 'CGCGCGGC', 'TGCGCAGC', 'TGCGCTGC', 'TGCGCGGC', 'AGCGCCAC', 'AGCGCCCC', 'AGCGCCTC', 'CGCGCCAC', 'CGCGCCCC', 'CGCGCCTC', 'TGCGCCAC', 'TGCGCCCC', 'TGCGCCTC', 'AGCGCCGA', 'AGCGCCGT', 'AGCGCCGG', 'CGCGCCGA', 'CGCGCCGT', 'CGCGCCGG', 'TGCGCCGA', 'TGCGCCGT', 'TGCGCCGG', 'GACGCCGC', 'GCCGCCGC', 'GTCGCCGC', 'GAAGCCGC', 'GATGCCGC', 'GAGGCCGC', 'GCAGCCGC', 'GCTGCCGC', 'GCGGCCGC', 'GTAGCCGC', 'GTTGCCGC', 'GTGGCCGC', 'GACACCGC', 'GACCCCGC', 'GACTCCGC', 'GCCACCGC', 'GCCCCCGC', 'GCCTCCGC', 'GTCACCGC', 'GTCCCCGC', 'GTCTCCGC', 'GACGACGC', 'GACGTCGC', 'GACGGCGC', 'GCCGACGC', 'GCCGTCGC', 'GCCGGCGC', 'GTCGACGC', 'GTCGTCGC', 'GTCGGCGC', 'GACGCAGC', 'GACGCTGC', 'GACGCGGC', 'GCCGCAGC', 'GCCGCTGC', 'GCCGCGGC', 'GTCGCAGC', 'GTCGCTGC', 'GTCGCGGC', 'GACGCCAC', 'GACGCCCC', 'GACGCCTC', 'GCCGCCAC', 'GCCGCCCC', 'GCCGCCTC', 'GTCGCCAC', 'GTCGCCCC', 'GTCGCCTC', 'GACGCCGA', 'GACGCCGT', 'GACGCCGG', 'GCCGCCGA', 'GCCGCCGT', 'GCCGCCGG', 'GTCGCCGA', 'GTCGCCGT', 'GTCGCCGG', 'GGAGCCGC', 'GGTGCCGC', 'GGGGCCGC', 'GGAACCGC', 'GGACCCGC', 'GGATCCGC', 'GGTACCGC', 'GGTCCCGC', 'GGTTCCGC', 'GGGACCGC', 'GGGCCCGC', 'GGGTCCGC', 'GGAGACGC', 'GGAGTCGC', 'GGAGGCGC', 'GGTGACGC', 'GGTGTCGC', 'GGTGGCGC', 'GGGGACGC', 'GGGGTCGC', 'GGGGGCGC', 'GGAGCAGC', 'GGAGCTGC', 'GGAGCGGC', 'GGTGCAGC', 'GGTGCTGC', 'GGTGCGGC', 'GGGGCAGC', 'GGGGCTGC', 'GGGGCGGC', 'GGAGCCAC', 'GGAGCCCC', 'GGAGCCTC', 'GGTGCCAC', 'GGTGCCCC', 'GGTGCCTC', 'GGGGCCAC', 'GGGGCCCC', 'GGGGCCTC', 'GGAGCCGA', 'GGAGCCGT', 'GGAGCCGG', 'GGTGCCGA', 'GGTGCCGT', 'GGTGCCGG', 'GGGGCCGA', 'GGGGCCGT', 'GGGGCCGG', 'GGCACCGC', 'GGCCCCGC', 'GGCTCCGC', 'GGCAACGC', 'GGCATCGC', 'GGCAGCGC', 'GGCCACGC', 'GGCCTCGC', 'GGCCGCGC', 'GGCTACGC', 'GGCTTCGC', 'GGCTGCGC', 'GGCACAGC', 'GGCACTGC', 'GGCACGGC', 'GGCCCAGC', 'GGCCCTGC', 'GGCCCGGC', 'GGCTCAGC', 'GGCTCTGC', 'GGCTCGGC', 'GGCACCAC', 'GGCACCCC', 'GGCACCTC', 'GGCCCCAC', 'GGCCCCCC', 'GGCCCCTC', 'GGCTCCAC', 'GGCTCCCC', 'GGCTCCTC', 'GGCACCGA', 'GGCACCGT', 'GGCACCGG', 'GGCCCCGA', 'GGCCCCGT', 'GGCCCCGG', 'GGCTCCGA', 'GGCTCCGT', 'GGCTCCGG', 'GGCGACGC', 'GGCGTCGC', 'GGCGGCGC', 'GGCGAAGC', 'GGCGATGC', 'GGCGAGGC', 'GGCGTAGC', 'GGCGTTGC', 'GGCGTGGC', 'GGCGGAGC', 'GGCGGTGC', 'GGCGGGGC', 'GGCGACAC', 'GGCGACCC', 'GGCGACTC', 'GGCGTCAC', 'GGCGTCCC', 'GGCGTCTC', 'GGCGGCAC', 'GGCGGCCC', 'GGCGGCTC', 'GGCGACGA', 'GGCGACGT', 'GGCGACGG', 'GGCGTCGA', 'GGCGTCGT', 'GGCGTCGG', 'GGCGGCGA', 'GGCGGCGT', 'GGCGGCGG', 'GGCGCAGC', 'GGCGCTGC', 'GGCGCGGC', 'GGCGCAAC', 'GGCGCACC', 'GGCGCATC', 'GGCGCTAC', 'GGCGCTCC', 'GGCGCTTC', 'GGCGCGAC', 'GGCGCGCC', 'GGCGCGTC', 'GGCGCAGA', 'GGCGCAGT', 'GGCGCAGG', 'GGCGCTGA', 'GGCGCTGT', 'GGCGCTGG', 'GGCGCGGA', 'GGCGCGGT', 'GGCGCGGG', 'GGCGCCAC', 'GGCGCCCC', 'GGCGCCTC', 'GGCGCCAA', 'GGCGCCAT', 'GGCGCCAG', 'GGCGCCCA', 'GGCGCCCT', 'GGCGCCCG', 'GGCGCCTA', 'GGCGCCTT', 'GGCGCCTG', 'GGCGCCGA', 'GGCGCCGT', 'GGCGCCGG']


molecule_count = {s: ((ds.sequence_subset == s.encode()) & (ds.sequence_name == b'HJ_general')).sum().item() for s in sequence_mutations_hj7}
mc = xr.DataArray(list(molecule_count.values()), dims=('sequence'), coords={'sequence': list(molecule_count.keys())})
mc.to_netcdf(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\HJ7 mutations\molecule_count.nc')


for seq in sequence_mutations_hj7:
    seq_sel = ((ds.sequence_subset == seq.encode()) & (ds.sequence_name == b'HJ_general'))
    trace_data = ds[['intensity_red_before', 'intensity', 'FRET']].sel(molecule=seq_sel)
    if len(trace_data.molecule) > 0:
        trace_data.to_netcdf(rf'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\Datasets per sequence\{seq}.nc')

from papylio.analysis.classification_simple import classify_anticorrelation



da = xr.DataArray(0, coords={'sequence': sequence_mutations_hj7 })

fig, ax = plt.subplots()




for seq in tqdm.tqdm(sequence_mutations_hj7):
    filepath = Path(rf'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\Datasets per sequence\{seq}.nc')
    if filepath.is_file():
        ds_seq = xr.load_dataset(filepath)
        ds_seq.close()
        ds_seq['selected'] = xr.DataArray(False, dims=('molecule',), coords={'molecule': ds_seq.molecule.values}).drop_vars('molecule')
        ds_seq['classification_anticorrelation'] = classify_anticorrelation(ds_seq.intensity, window=20, rolling_mean_window=10, threshold=-0.5)
        ds_seq.to_netcdf(rf'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\Datasets per sequence\{seq}.nc')


counts = []
anticorrelation_values = []
for seq in tqdm.tqdm(sequence_mutations_hj7):
    filepath = Path(rf'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\Datasets per sequence\{seq}.nc')
    if filepath.is_file():
        ds_seq = xr.load_dataset(filepath)
        anticorrelation_with_red = ds_seq['classification_anticorrelation'].sel(molecule=(ds_seq.intensity_red_before.sel(channel=1)>2000))
        if len(anticorrelation_with_red.molecule)>0:
            anticorrelation_value = (anticorrelation_with_red.sum() / anticorrelation_with_red.size).item()
            anticorrelation_values.append(anticorrelation_value)
            counts.append(len(anticorrelation_with_red.molecule))
        else:
            anticorrelation_values.append(np.nan)
            counts.append(0)
            print('No red')
    else:
        anticorrelation_values.append(np.nan)
        counts.append(0)

cc = xr.DataArray(counts, dims=('sequence'), coords={'sequence': sequence_mutations_hj7}, name='Molecule count')
ac = xr.DataArray(anticorrelation_values, dims=('sequence'), coords={'sequence': sequence_mutations_hj7}, name='Fraction showing anticorrelation')
eds = ac.to_dataset()
eds['Molecule count'] = cc
eds.to_netcdf(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\HJ7 mutations\fraction_anticorrelation.nc')


from papylio.trace_plot import TracePlotWindow
from papylio.experiment import get_QApplication
app = get_QApplication()
frame = TracePlotWindow(ds_seq, plot_variables=['intensity', 'FRET', 'classification_anticorrelation'],
                        ylims=[(0, 15000), (-0.1,1.1), (-1,2)], colours=[('g', 'r'), ('b'), ('k')], save_path=None)
app.exec_()







for seq in sequence_mutations_hj7:
    seq_sel = ((ds.sequence_variable == seq.encode()))
    trace_data = ds[['intensity_red_laser', 'intensity', 'FRET', 'classification']].sel(molecule=seq_sel)
    if len(trace_data.molecule) > 0:
        trace_data.to_netcdf(exp.main_path.joinpath('Analysis').joinpath('HJ7 mutations').joinpath(seq).with_suffix('.nc'))


counts = []
anticorrelation_values = []
for seq in tqdm.tqdm(sequence_mutations_hj7):
    filepath = exp.main_path.joinpath('Analysis').joinpath('HJ7 mutations').joinpath(seq).with_suffix('.nc')
    if filepath.is_file():
        ds_seq = xr.load_dataset(filepath)
        anticorrelation_trace = ds_seq['classification']>=0
        if len(anticorrelation_trace.molecule)>0:
            anticorrelation_value = (anticorrelation_trace.sum() / anticorrelation_trace.size).item()
            anticorrelation_values.append(anticorrelation_value)
            counts.append(len(anticorrelation_trace.molecule))
        else:
            anticorrelation_values.append(np.nan)
            counts.append(0)
    else:
        anticorrelation_values.append(np.nan)
        counts.append(0)


cc = xr.DataArray(counts, dims=('sequence'), coords={'sequence': sequence_mutations_hj7}, name='Molecule count')
ac = xr.DataArray(anticorrelation_values, dims=('sequence'), coords={'sequence': sequence_mutations_hj7}, name='Fraction showing anticorrelation')
eds = ac.to_dataset()
eds['Molecule count'] = cc
eds.to_netcdf(exp.main_path.joinpath('Analysis').joinpath('HJ7 mutations').joinpath('fraction_anticorrelation.nc'))
# eds.to_netcdf(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\HJ7 mutations\fraction_anticorrelation.nc')


# ds = xr.open_mfdataset([file.relativeFilePath.with_suffix('.nc') for file in files_green_laser[0:300]], combine='nested', concat_dim='molecule',
#                        data_vars='minimal', coords='minimal', compat='override', engine='h5netcdf')#, parallel=True)#, chunks={'molecule': 100})
import time
start = time.time()
ds = xr.open_mfdataset([file.relativeFilePath.with_suffix('.nc') for file in files_green_laser], combine='nested', concat_dim='molecule',
                       data_vars='minimal', coords='minimal', compat='override', engine='h5netcdf', parallel=False)
print(time.time()-start)

import time
start = time.time()
ds.to_netcdf(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20211005 - Objective-type TIRF (BN)\complete_dataset.nc', engine='h5netcdf')
print(time.time()-start)

def read_netcdfs(files, dim, transform_func=None):
    paths = files #sorted(glob(files))
    datasets = []
    for path in paths:
        with xr.open_dataset(path, engine='h5netcdf') as ds:
            if transform_func is not None:
                ds = transform_func(ds)
            if len(ds.molecule)>0:
                ds.load()
                datasets.append(ds)
    combined = xr.concat(datasets, dim)
    return combined

def ff(ds):
    return ds.sel(molecule=(ds.sequence_variable == 'CAGCAGCA'))

filepaths = files_green_laser.absoluteFilePath.with_suffix('.nc')
start = time.time()
test = read_netcdfs(filepaths, 'molecule', transform_func=ff)
print(time.time()-start)


for file in tqdm.tqdm(files_green_laser[400:]):
    ds = file.dataset

    # s = ds['sequence']==''
    # ds['sequence'][s]='-'*120
    # ds['sequence_quality'][s]=' '*120
    # ds = ds.drop_vars('distance_to_sequence', errors='ignore')
    # for i, index in enumerate([30,31,56,57,82,83,108,109]):
    #     if i == 0:
    #         sequence_variable = ds.sequence.str.get(index)
    #     else:
    #         sequence_variable += ds.sequence.str.get(index)
    #
    # ds['sequence_variable'] = sequence_variable

    # keys = ['file','dimension','sequence','sequence_quality','sequence_variable']
    # for key in keys:
    #     ds[key] = ds[key].astype('|S')

    encoding = {'file': {'dtype': '|S'}, 'dimension': {'dtype': '|S'}, 'selected': {'dtype': bool}}
    ds.to_netcdf(file.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='w', encoding=encoding)


def change_type(file):
    ds = file.dataset
    encoding = {'file': {'dtype': '|S'}, 'dimension': {'dtype': '|S'}, 'selected': {'dtype': bool}}
    ds.to_netcdf(file.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='w', encoding=encoding)

import joblib
from objectlist.base import tqdm_joblib
# for file_green, file_red_before, file_red_after in
with tqdm_joblib(tqdm.tqdm(total=len(files_green_laser[400:]))):
    joblib.parallel.Parallel(6)(joblib.parallel.delayed(change_type)(file) for file in files_green_laser[400:])

# -----------------------------------
# Select sequencing data for mapping
# -----------------------------------
mapping_sequence = 'ACTGACTGTAACAACAACAACAATAACAACAACAACAATAACAACAACAACAAT'
number_of_allowed_mismatches = 2
exp.select_sequencing_data_for_mapping(mapping_sequence, number_of_allowed_mismatches)

# -----------------------------------
# Find high-intensity coordinates
# -----------------------------------
configuration = exp.configuration['find_coordinates'].copy()
configuration['peak_finding']['minimum_intensity_difference'] = 1500

for file in files_red_laser:
    file.find_coordinates(configuration=configuration)


# -----------------------------------
# Geometric hashing 2.0
# -----------------------------------
# # This part is likely not yet working well
# start = time.time()
# exp.generate_mapping_hashtable(mapping_sequence, number_of_allowed_mismatches,
#                                imaged_surface=2, maximum_distance_tile=3000, tuple_size=4)
# end = time.time()
# print(f'Hashtable generation: {end - start} seconds')
#
# # Or just use the coordinates from the loc files
# # path_to_loc_folder = Path(r'Path_to_loc_folder')
# # tile_coordinate_files = ['2101.loc', '2102.loc', '2103.loc', '2104.loc']
# # tile_coordinate_sets = [np.loadtxt(path_to_loc_folder.joinpath(f)) for f in tile_coordinate_files]
# # start = time.time()
# # exp.generate_mapping_hashtable_from_coordinate_set(tile_coordinate_sets, maximum_distance_tile=3000, tuple_size=4)
# # end = time.time()
# # print(f'Hashtable generation: {end - start} seconds')
#
# # For each file in experiment find a match
# # NOTE: 'scale': [-1,1] means a reflection with respect to the y axis.
#
# start = time.time()
# for file in exp.files:
#     file.sequencing_match = None
#     file.find_sequences(maximum_distance_file=1000, tuple_size=4, initial_transformation={'scale': [-1,1]},
#                         hash_table_distance_threshold=0.01,
#                         alpha=0.1, test_radius=10, K_threshold=10e6, # original K_threshold = 10e9
#                         magnification_range=[3.3,3.4], rotation_range=[-1,1])
#     print(file)
# end = time.time()
# print(f'Matching: {end - start} seconds')
#
# matched_files = [file for file in exp.files if file.sequencing_match]
# number_of_matches = len(matched_files)
#
# # Improve mapping by performing a linear least-squares fit on all nearest neighbours within the distance threshold
# for file in matched_files:
#     nearest_neighbour_match_distance_threshold = 25
#     file.sequencing_match.nearest_neighbour_match(nearest_neighbour_match_distance_threshold)
#
# # Make and export a sequencing match plot for files that have a sequencing match
# for file in matched_files: #exp.files # files_correct: #exp.files:
#     if file.sequencing_match:
#         file.plot_sequencing_match()
#         # plt.close()

# -----------------------------------
# ASSESSING OVERALL MAPPING QUALITY
# -----------------------------------
#
# # Calculate rotation and magnification for all found matches
# rotation = np.array([file.sequencing_match.rotation for file in exp.files if file.sequencing_match])
# magnification = np.array([file.sequencing_match.magnification for file in exp.files if file.sequencing_match])
# mean_magnification = np.mean(magnification, axis=1)
#
# # Make a heatmap scatter plot
# from scipy.stats import gaussian_kde
# x = rotation
# y = mean_magnification
#
# xy = np.vstack([x,y])
# z = gaussian_kde(xy,0.01)(xy)
# fig, ax = plt.subplots()
# ax.scatter(x, y, c=z, s=100, edgecolor='')
# plt.title('Matches')
# plt.xlabel('Rotation (deg)')
# plt.ylabel('Magnfication (average over two directions)')
# plt.show()
#
# # Make a heatmap scatter plot
# from scipy.stats import gaussian_kde
# selection = ((rotation>179) | (rotation<-179)) & (mean_magnification>3.3) & (mean_magnification<3.4)
# x = rotation[selection]
# y = mean_magnification[selection]
# xy = np.vstack([x,y])
# z = gaussian_kde(xy,0.01)(xy)
# fig, ax = plt.subplots()
# ax.scatter(x, y, c=z, s=100, edgecolor='')
# plt.title('Matches')
# plt.xlabel('Rotation (deg)')
# plt.ylabel('Magnfication (average over two directions)')
# plt.show()


# -----------------------------------
# Geometric hashing 3.0
# -----------------------------------

# Define initial file transformation
initial_magnification = np.array([3.71, -3.71])
initial_rotation = 0.4

initial_file_transformation = AffineTransform(matrix=None, scale=initial_magnification,
                                                rotation=initial_rotation/360*np.pi*2,
                                                shear=None, translation=None)

# TODO: Add timer to generate_mapping_hashtable and find_sequences methods, by making a decorator function. [IS: 10-08-2020]
start = time.time()
exp.generate_mapping_hashtable3(mapping_sequence, number_of_allowed_mismatches,
                               imaged_surface=1, initial_file_transformation=initial_file_transformation)
end = time.time()
print(f'Hashtable generation: {end - start} seconds')

matched_files = []
start = time.time()
for file in files_red_laser:
    if (file.number_of_molecules>4):
        print(file)
        file.sequencing_match = None
        # file.find_sequences3(distance=15, alpha=0.8, sigma=5, K_threshold=10**7)
        file.find_sequences3(distance=10, alpha=0.9, sigma=5, K_threshold=10**9, channel=1)
        if file.sequencing_match:
            matched_files.append(file)
end = time.time()
print(f'Matching: {end - start} seconds')

for file in matched_files:
    print(file.name + ' __ ' + str(file.sequencing_match.tile))

# Improve mapping by performing a linear least-squares fit on all nearest neighbours within the distance threshold
for file in matched_files:
    print(file)
    # try:
    nearest_neighbour_match_distance_threshold = 25
    file.sequencing_match.nearest_neighbour_match(nearest_neighbour_match_distance_threshold, transformation_type='linear')
    # except:
    #     print(file)

exp.show_sequencing_matches()

for file in matched_files:
    file.plot_sequencing_match()
    file.export_sequencing_match()


# -----------------------------------
# Find sequencing matches using stage coordinates
# -----------------------------------
# Find mapping between stage coordinates and sequencing coordinates
exp.map_sequencing_and_stage_coordinates()
exp.show_stage_to_sequencing_mappings()

# Find sequencing matches using stage coordinates
for file in files_red_laser:
    file.find_sequences_using_stage_coordinates(channel=1, show=True, save=False)

exp.show_sequencing_matches()

# Improve mapping by performing a linear least-squares fit on all nearest neighbours within the distance threshold
for file in files_red_laser:
    print(file)
    file.sequencing_match.nearest_neighbour_match(25, 'linear')
    file.sequencing_match.nearest_neighbour_match(10, 'linear')
    # file.sequencing_match.nearest_neighbour_match(10, 'polynomial', order=4)

file_red.sequencing_match.show_mapping_transformation(inverse=True, crop=True)


# -----------------------------------
# Classify traces in sequencing data of file
# -----------------------------------
# Doing this on all sequencing data is still slow.




# -*- coding: utf-8 -*-
"""
Created on Tue Dec 13 19:25:17 2016

@author: ikatechis
"""

import numpy as np
import matplotlib.pyplot as plt
from os import listdir
import os


#@timer
def stepfinder(trace, threshold=100, max_steps=20):

    start_frames = []
    if trace[0] > threshold and trace[1] > threshold:
        start_frames.append(0)

    stop_frames = []
    if trace[-1] > threshold and trace[-2] > threshold:
        stop_frames.append(trace.size - 1)

    i = 0
    while i < trace.size -2:
        print(i)
        dif1 = trace[i+1] - trace[i]
        dif2 = trace[i+2] - trace[i]

        if ((dif1 > threshold and
            dif2 > threshold and
            trace[i] < threshold) or
            (i==0 and start_frames) ): # this will not catch stoichiometry of 2
            # start_frames.append(i+1)
            for j in range(i+2, trace.size - 2):  # start 2 positions after the step start until the length of the original trace
                dif1 = trace[j+1] - trace[j]
                dif2 = trace[j+2] - trace[j]

                if dif1 < -threshold and dif2 < -threshold\
                            and trace[j+2] < threshold:
                    start_frames.append(i+1)
                    stop_frames.append(j+1)
                    i = j+1
                    break

        i +=1
        # i += 1 # start the next loop from the last stop frame

    if len(start_frames) != len(stop_frames):  # sometimes 2 consecutive frames both satisfy the threshold condition for very big jumps
        start_temp = np.copy(start_frames)
        stop_temp = np.copy(stop_frames)
        for i, start in enumerate(start_temp):
            if  start_temp[i] - start_temp[i-1] == 1:
#                print('Found two consecutive start frames')

                start_frames.remove(start)
        for i, stop in enumerate(stop_temp):
            if  stop_temp[i] - stop_temp[i-1] == 1:
                print('Found two consecutive stop frames')
                stop_frames.remove(stop)

    if len(start_frames) != len(stop_frames): # if the problem remains
        print ("something is wrong")
        print ("start frames: "+str(start_frames))
        print ("stop frames: "+str(stop_frames))
        return {"frames": np.array([])}

    elif len(start_frames) > max_steps:
        print ("Found more steps than the limit of "+str(max_steps))
        return {"frames": np.array([])}
    else:
        print ("steps found: " + str(len(start_frames+stop_frames)))
        if len(start_frames+stop_frames) % 2 > 0:
            print('odd number of steps found. Result discarded.')
        print ("start frames: "+str(start_frames))
        print ("stop frames: "+str(stop_frames))
        res={ "frames": np.array(start_frames+stop_frames),
             "threshold": threshold}
        return res


def plot_steps(trace="red", exposure_time=0.1, steps={},
               name="molecule_0", display_plot=False,
               save_plot=True, save_folder="./saved_plots"):

    Nframes = trace.size
    time = np.linspace(0, Nframes*exposure_time, Nframes)
    if not display_plot:
        plt.ioff()
    plt.figure(name, figsize=(10, 4))


    if "start_times" in steps.keys():
        start_times = steps["start_times"]
        stop_times = steps["stop_times"]
        for start, stop in zip(start_times, stop_times):
            plt.axvline(start, c="green", lw=2, ls="-.")
            plt.axvline(stop, c="red", lw=1, ls="--")
    plt.plot(time, trace, "k", lw=2)
    plt.xlabel("Time (s)")
    plt.ylabel("Counts")
    plt.xlim((min(time), max(time)))
    plt.tight_layout()
    if save_plot:
        if save_folder not in listdir("."):
            os.mkdir(save_folder)
        plt.savefig("./"+save_folder+"/"+name)
        if not display_plot:
            plt.close()
            plt.pause(0.001)
    plt.ion()


import pytest
import numpy as np
from papylio.netcdf_operations import merge_datasets, reorder_datasets_using_sequence_subset
import netCDF4

@pytest.fixture
def netcdf_filepaths(shared_datadir):
    return list((shared_datadir / 'netcdf_operations').glob('*.nc'))


def test_merge_datasets(shared_datadir, netcdf_filepaths):
    merge_datasets(netcdf_filepaths, shared_datadir / 'file_out.nc', 'molecule', init_file=None, with_sequence_only=False)

    ds_out = netCDF4.Dataset(shared_datadir / 'file_out.nc')
    assert ds_out.dimensions['molecule'].size == np.sum([netCDF4.Dataset(p).dimensions['molecule'].size for p in netcdf_filepaths])
    assert (ds_out['intensity'][1707:3331] == netCDF4.Dataset(netcdf_filepaths[1])['intensity'][:]).all()


def test_merge_datasets_sequence_only(shared_datadir, netcdf_filepaths):
    merge_datasets(netcdf_filepaths, shared_datadir / 'file_out.nc', 'molecule', init_file=None, with_sequence_only=True)

    ds_out = netCDF4.Dataset(shared_datadir / 'file_out.nc')
    assert ds_out.dimensions['molecule'].size == np.sum([(netCDF4.Dataset(p)['sequence'][:]!=b'-').all(axis=1).sum() for p in netcdf_filepaths])
    ds_in1 = netCDF4.Dataset(netcdf_filepaths[1])
    assert (ds_out['intensity'][319:319+318] == ds_in1['intensity'][(ds_in1['sequence'][:]!=b'-').all(axis=1)]).all()

def test_reorder_datasets_using_sequence_subset(shared_datadir, netcdf_filepaths):
    reorder_datasets_using_sequence_subset(netcdf_filepaths, shared_datadir, 'molecule')




# def test_merge_datasets():
#     from pathlib2 import Path
#     shared_datadir = Path(r'H:\My Documents\Python\traceanalysis\tests\integration\data\netcdf_operations')
#     netcdf_filepaths = list(shared_datadir.glob('*.nc'))
#     merge_datasets(netcdf_filepaths, shared_datadir / 'file_out.nc', 'molecule', init_file=None, with_sequence_only=False)
#     assert 1 == 1



import sys
import time

import matplotlib.pyplot as plt
import numpy as np

from papylio.file import calculate_FRET, calculate_stoichiometry, calculate_intensity_total

from PySide2 import QtWidgets
from PySide2.QtWidgets import QMainWindow, QPushButton, QWidget, QVBoxLayout, QLineEdit, QLabel
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from matplotlib.backends.backend_qt5agg import (
    FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure, GridSpec

import sys
import time

import numpy as np

# from matplotlib.backends.qt_compat import QtWidgets
# from matplotlib.backends.backend_qtagg import (
#     FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
# from matplotlib.figure import Figure

def trace_correction(intensity, background_correction=None, alpha_correction=None,
                     gamma_correction=None):
    intensity = intensity.copy()
    if background_correction is not None:
        intensity[dict(channel=0)] -= background_correction[0]
        intensity[dict(channel=1)] -= background_correction[1]
        intensity.attrs['background_correction'] = background_correction
    if gamma_correction is not None:
        intensity[dict(channel=0)] *= gamma_correction
        intensity.attrs['gamma_correction'] = gamma_correction
    if alpha_correction is not None:
        intensity[dict(channel=0)] += alpha_correction * intensity[dict(channel=0)]
        intensity[dict(channel=1)] -= alpha_correction * intensity[dict(channel=0)]
        intensity.attrs['alpha_correction'] = alpha_correction
    # if delta_correction is not None:
    #     intensity[dict(channel=0)] *= self.delta_correction

    # if beta_correction is not None:
    #     intensity[dict(channel=0)] *= self.beta_correction
    return intensity


# if beta_correction is not None:
#     intensity[dict(channel=0)] *= self.beta_correction


class TraceCorrectionWindow(QWidget):
    def __init__(self, intensity_raw, show=True):
        from papylio.experiment import get_QApplication
        app = get_QApplication()

        super().__init__()

        self.setWindowTitle("Trace correction")

        self.setFixedHeight(600)
        self.setFixedWidth(1500)

        self.intensity_raw = intensity_raw #.values.flatten()
        self.background_correction_green = 0
        self.background_correction_red = 0
        self.alpha_correction = 0
        self.gamma_correction = 1


        self.histogram_intensity_green_intensity_red_kwargs = \
            dict(bins=(1000,1000))

        self.histogram_FRET_intensity_total_kwargs = \
            dict(bins=(1000,1000), range=((-0.05,1.05),(-np.max(self.intensity_total)*0.05,np.max(self.intensity_total))))

        self.histogram_FRET_stoichiometry_kwargs = \
            dict(bins=(1000,1000), range=((-0.05,1.05),(-0.05,1.05)))


        illumination_indices = np.unique(self.intensity_raw.illumination)

        if len(illumination_indices) == 1:
            figure = plt.figure(figsize=(9, 6), layout='constrained')
            self.gs = figure.add_gridspec(1, 2)
        elif len(illumination_indices) == 2:
            figure = plt.figure(figsize=(14, 6), layout='constrained')
            self.gs = figure.add_gridspec(1, 3)
        else:
            raise NotImplementedError('More than two illuminations are not yet supported, feel free to add this feature yourself!')

        # figure = plt.figure(figsize=(14, 6), layout='constrained')
        # self.gs = figure.add_gridspec(1, 3)

        self.axis_G_R = figure.add_subplot(self.gs[0,0])
        self.axis_G_R.set_box_aspect(1)
        self.axis_G_R.set_aspect(1)

        from mpl_toolkits.axes_grid1 import make_axes_locatable
        self.axis_E_T = figure.add_subplot(self.gs[0,1])
        self.axis_E_T.set_box_aspect(1)
        # divider_E_T = make_axes_locatable(self.axis_E_T)
        # self.axis_E = divider_E_T.append_axes("top", size="25%", pad=0)
        # self.axis_T = divider_E_T.append_axes("right", size="25%", pad=0)


        if len(illumination_indices) == 2:
            self.axis_E_S = figure.add_subplot(self.gs[0, 2], sharex=self.axis_E_T)
            self.axis_E_S.set_box_aspect(1)
            # divider_E_S = make_axes_locatable(self.axis_E_S)
            # self.axis_S = divider_E_S.append_axes("right", size="25%", pad=0)
        else:
            self.axis_E_S = None
            # self.axis_S = None

        # for axis in [self.ax]

        self.canvas = FigureCanvas(figure)
        # Ideally one would use self.addToolBar here, but it is slightly
        # incompatible between PyQt6 and other bindings, so we just add the
        # toolbar as a plain widget instead.
        figure_layout = QtWidgets.QVBoxLayout()
        figure_layout.addWidget(NavigationToolbar(self.canvas, self))
        figure_layout.addWidget(self.canvas)

        control_layout = QtWidgets.QGridLayout()
        corrections = ['background_correction_green', 'background_correction_red', 'alpha_correction', 'gamma_correction']
        self.text_boxes = {}
        for i, correction in enumerate(corrections):
            label = control_layout.addWidget(QLabel(correction), i, 0)
            line_edit = QLineEdit(str(getattr(self, correction)))
            # line_edit.textChanged.connect(lambda text: setattr(self, correction, float(text)))
            self.text_boxes[correction] = line_edit
            line_edit.editingFinished.connect(self.setCorrectionsFromTextBoxes)
            # line_edit.editingFinished.connect(self.plot_histogram_FRET_intensity_total)
            control_layout.addWidget(line_edit, i, 1)


        main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(main_layout)
        main_layout.addLayout(figure_layout)
        main_layout.addLayout(control_layout)

        # dynamic_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        # layout.addWidget(dynamic_canvas)
        # layout.addWidget(NavigationToolbar(dynamic_canvas, self))




        # self.axis.hist2d(np.zeros((self.FRET.size, self.intensity_total.size)),
        #                  cmin=1)
        # H, x, y = self.histogram_FRET_intensity_total()
        # H[H==0] = np.nan
        # self.axis.imshow(H, origin='lower')
        # self.axis.pcolormesh()
        # fig, ax = plt.subplots()
        # ax.hist2d(self.FRET.values.flatten(), self.intensity_total.values.flatten(), cmin=1, **self.histogram_FRET_intensity_total_kwargs)
        # plt.show()
        # self.plot_histogram_FRET_intensity_total()
        # self.axis.set_xlabel('FRET')
        # self.axis.set_ylabel('Intensity total')
        # self._dynamic_ax = dynamic_canvas.figure.subplots()
        # t = np.linspace(0, 10, 101)
        # # Set up a Line2D.
        # self._line, = self._dynamic_ax.plot(t, np.sin(t + time.time()))
        # self._timer = dynamic_canvas.new_timer(50)
        # self._timer.add_callback(self._update_canvas)
        # self._timer.start()

        if show:
            self.show()
            self.activateWindow()
            self.raise_()
            app.exec_()

        self.update_canvas()

    @property
    def intensity(self):
        return trace_correction(self.intensity_raw, (self.background_correction_green, self.background_correction_red),
                                self.alpha_correction, self.gamma_correction)

    @property
    def intensity_total(self):
        return calculate_intensity_total(self.intensity).sel(frame=self.intensity.illumination==0)

    @property
    def FRET(self):
        return calculate_FRET(self.intensity).sel(frame=self.intensity.illumination==0)

    @property
    def stoichiometry(self):
        return calculate_stoichiometry(self.intensity)

    # def histogram_FRET_intensity_total(self):
    #     FRET = self.FRET.values.flatten()
    #     intensity_total = self.intensity_total.values.flatten()
    #     H, xedges, yedges = np.histogram2d(FRET, intensity_total, **self.histogram_FRET_intensity_total_kwargs)
    #     xcenters = (xedges[1:] + xedges[:-1])/2
    #     ycenters = (yedges[1:] + yedges[:-1]) / 2
    #     return H.T, xcenters, ycenters

    def setCorrectionsFromTextBoxes(self):
        for correction, line_edit in self.text_boxes.items():
            setattr(self, correction, float(line_edit.text()))
            print(float(line_edit.text()))

        self.update_canvas()

    def plot_histogram_intensity_green_intensity_red(self):
        print(self.background_correction_green, self.background_correction_red, self.alpha_correction, self.gamma_correction)

        axis = self.axis_G_R

        intensity = self.intensity

        axis.cla()
        intensity_green = intensity.sel(channel=0, frame=intensity.illumination==0).values.flatten()
        intensity_red = intensity.sel(channel=1, frame=intensity.illumination==0).values.flatten()
        axis.hist2d(intensity_green, intensity_red, cmin=1, **self.histogram_intensity_green_intensity_red_kwargs)
        axis.set_xlabel('Intensity green')
        axis.set_ylabel('Intensity red')
        (xmin, xmax) = axis.get_xlim()
        (ymin, ymax) = axis.get_ylim()
        axis.hlines(y=0, xmin=xmin, xmax=xmax)
        axis.vlines(x=0, ymin=ymin, ymax=ymax)
        axis.set_xlim(xmin, xmax)
        axis.set_ylim(ymin, ymax)


    def plot_histogram_FRET_intensity_total(self):
        print(self.background_correction_green, self.background_correction_red, self.alpha_correction, self.gamma_correction)

        axis = self.axis_E_T

        axis.cla()
        FRET = self.FRET.values.flatten()
        intensity_total = self.intensity_total.values.flatten()
        axis.hist2d(FRET, intensity_total, cmin=1, **self.histogram_FRET_intensity_total_kwargs)
        axis.set_xlabel('FRET')
        axis.set_ylabel('Intensity total')
        (xmin, xmax), (ymin, ymax) = self.histogram_FRET_intensity_total_kwargs['range']
        axis.hlines(y=0, xmin=xmin, xmax=xmax)
        axis.vlines(x=[0,0.5,1], ymin=ymin, ymax=ymax)
        axis.set_xlim(xmin, xmax)
        axis.set_ylim(ymin, ymax)



    def plot_histogram_FRET_stoichiometry(self):

        axis = self.axis_E_S

        axis.cla()
        FRET = self.FRET.values.flatten()
        stoichiometry = self.stoichiometry.values.flatten()
        axis.hist2d(FRET, stoichiometry, cmin=1, **self.histogram_FRET_stoichiometry_kwargs)
        axis.set_xlabel('FRET')
        axis.set_ylabel('Stoichiometry')
        axis.figure.canvas.draw()
        (xmin, xmax), (ymin, ymax) = self.histogram_FRET_stoichiometry_kwargs['range']
        axis.hlines(y=[0,0.5,1], xmin=xmin, xmax=xmax)
        axis.vlines(x=[0,0.5,1], ymin=ymin, ymax=ymax)
        axis.set_xlim(xmin, xmax)
        axis.set_ylim(ymin, ymax)


    def update_canvas(self):
        self.plot_histogram_intensity_green_intensity_red()
        self.plot_histogram_FRET_intensity_total()
        if self.axis_E_S is not None:
            self.plot_histogram_FRET_stoichiometry()
        self.canvas.draw()

        # t = np.linspace(0, 10, 101)
        # # Shift the sinusoid as a function of time.
        # self._line.set_data(t, np.sin(t + time.time()))
        # self._line.figure.canvas.draw()

#
# class TraceCorrectionWindow(QWidget):
#     def __init__(self, traces=None, parent=None, show=True):
#         from papylio.experiment import get_QApplication
#         app = get_QApplication()
#
#         super().__init__()
#         self.parent = parent
#
#         self.traces = traces
#
#
#
#         if show:
#             self.show()
#             self.activateWindow()
#             self.raise_()
#             app.exec_()
#
#         self.setWindowTitle("Trace correction")
#
#         layout = QVBoxLayout(self)
#
#         static_canvas = FigureCanvasQTAgg(Figure(figsize=(5, 3)))
#         # Ideally one would use self.addToolBar here, but it is slightly
#         # incompatible between PyQt6 and other bindings, so we just add the
#         # toolbar as a plain widget instead.
#         layout.addWidget(NavigationToolbar(static_canvas, self))
#         layout.addWidget(static_canvas)
#
#         static_canvas.draw()
#         dynamic_canvas = FigureCanvasQTAgg(Figure(figsize=(5, 3)))
#         layout.addWidget(dynamic_canvas)
#         layout.addWidget(NavigationToolbar(dynamic_canvas, self))
#
#         self._static_ax = static_canvas.figure.subplots()
#         t = np.linspace(0, 10, 501)
#         self._static_ax.plot(t, np.tan(t), ".")
#
#         self._dynamic_ax = dynamic_canvas.figure.subplots()
#         t = np.linspace(0, 10, 101)
#         # Set up a Line2D.
#         self._line, = self._dynamic_ax.plot(t, np.sin(t + time.time()))
#         self._timer = dynamic_canvas.new_timer(50)
#         self._timer.add_callback(self._update_canvas)
#         self._timer.start()
#
#     def _update_canvas(self):
#         t = np.linspace(0, 10, 101)
#         # Shift the sinusoid as a function of time.
#         self._line.set_data(t, np.sin(t + time.time()))
#         self._line.figure.canvas.draw()
#
#
# if __name__ == "__main__":
#     w = TraceCorrectionWindow()



import numpy as np #scientific computing with Python
import xarray as xr
import matplotlib.pyplot as plt

# import matplotlib.patches as patches
# from matplotlib.collections import PatchCollection
# from mpl_toolkits.axes_grid1 import make_axes_locatable

# from papylio.molecule import Molecule


def marginal_hist2d(x, y, bins=10, range=(0, 1), xlabel=None, ylabel=None, count_label='Count', show_marginal=True,
                    show_colorbar=False, ax=None, **hist2d_kwargs):
    """
    Generates a 2D histogram plot with optional marginal histograms and colorbar.

    Parameters:
    -----------
    x : array-like or xarray.DataArray
        The data for the x-axis. If an xarray.DataArray is provided, its name will be used as the x-axis label.
    y : array-like or xarray.DataArray
        The data for the y-axis. If an xarray.DataArray is provided, its name will be used as the y-axis label.
    bins : int or tuple of int, optional (default=100)
        Number of bins for the 2D histogram. If an integer is provided, it will be used for both x and y axes.
    range : tuple or tuple of two tuples, optional (default=((0, 1), (0, 5000)))
        The range of the data for the histogram in the form (min, max) or ((xmin, xmax), (ymin, ymax)).
    ax : matplotlib.axes.Axes, optional (default=None)
        The axes object to plot on. If None, a new figure and axes will be created.
    xlabel : str, optional (default=None)
        Label for the x-axis. If not provided, the name of the `x` data will be used (if available).
    ylabel : str, optional (default=None)
        Label for the y-axis. If not provided, the name of the `y` data will be used (if available).
    show_marginal : bool, optional (default=True)
        Whether to show the marginal histograms along the x and y axes.
    show_colorbar : bool, optional (default=True)
        Whether to display a colorbar next to the 2D histogram.
    hist2d_kwargs : dict, optional (default=None)
        Additional keyword arguments passed to `matplotlib.pyplot.hist2d` for customizing the 2D histogram.
    count_label : str, optional (default='Counts')
        Label for the count axis (used in the marginal histograms and colorbar).

    Returns:
    --------
    fig : matplotlib.figure.Figure
        The figure object containing the plot.
    axs : list of matplotlib.axes.Axes
        A list containing the axes objects for the main plot, marginal histograms (if shown), and colorbar (if shown).

    """

    if ax is None:
        fig, ax_main = plt.subplots(figsize=(5, 5))
    else:
        fig = ax.figure
        ax_main = ax

    if isinstance(x, xr.DataArray):
        if xlabel is None:
            xlabel = x.name
        x = x.values.flatten()

    if isinstance(y, xr.DataArray):
        if ylabel is None:
            ylabel = y.name
        y = y.values.flatten()

    if isinstance(bins, int):
        bins = [bins] * 2

    if np.issubdtype(type(range[0]), np.number):
        range = (range,) * 2

    # Main 2D histogram
    h = ax_main.hist2d(x, y, bins=bins, range=range, **hist2d_kwargs)
    ax_main.set_box_aspect(1)

    ax_main.set_xlabel(xlabel)
    ax_main.set_ylabel(ylabel)

    axs = [ax_main]

    # Marginal histograms
    if show_marginal == True:
        cmap = plt.get_cmap('viridis')
        marginal_hist_kwargs = dict(fc=cmap(0))

        spacing = 0.01
        xyhist_size = 0.15
        bbox_main = ax_main.get_position()

        ax_xhist = fig.add_axes(
            rect=[bbox_main.x0, bbox_main.y0 + bbox_main.height + spacing, bbox_main.width, xyhist_size],
            sharex=ax_main)
        ax_yhist = fig.add_axes(
            rect=[bbox_main.x0 + bbox_main.width + spacing, bbox_main.y0, xyhist_size, bbox_main.height],
            sharey=ax_main)

        ax_xhist.hist(x, bins=bins[0], range=range[0], **marginal_hist_kwargs)
        ax_yhist.hist(y, bins=bins[1], range=range[1], orientation='horizontal', **marginal_hist_kwargs)

        # Hide the tick labels on the marginal histograms
        ax_xhist.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
        ax_yhist.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)

        # Set custom ticks excluding 0
        x_hist_y_ticks = ax_xhist.get_yticks()
        y_hist_x_ticks = ax_yhist.get_xticks()
        ax_xhist.set_yticks(x_hist_y_ticks[1:])
        ax_yhist.set_xticks(y_hist_x_ticks[1:])

        # Set labels
        ax_xhist.set_ylabel(count_label)
        ax_yhist.set_xlabel(count_label)

        axs.append(ax_xhist)
        axs.append(ax_yhist)

    # Create a colorbar axis that matches the height of the main plot
    if show_colorbar:
        bbox_yhist = ax_yhist.get_position()

        ax_cbar = fig.add_axes(
            rect=[bbox_main.x0 + bbox_main.width + bbox_yhist.width + spacing * 2, bbox_main.y0, xyhist_size * 0.2,
                  bbox_main.height])
        fig.colorbar(h[3], cax=ax_cbar)
        ax_cbar.set_ylabel(count_label)

        axs.append(ax_cbar)

    return fig, axs


def histogram(da, axis=None, **hist_kwargs):
    if axis is None:
        figure, axis = plt.subplots()
    else:
        figure = axis.figure

    if 'channel' in da.dims:
        das = [da.sel(channel=channel) for channel in da.channel]
    else:
        das = [da]

    if len(das) > 1:
        hist_kwargs['histtype'] = 'step'

    for da in das:
        da.plot.hist(ax=axis, **hist_kwargs)
    axis.set_ylabel('Count')
    axis.set_title('')

    # if save:
    #     fig.savefig(self.absoluteFilePath.with_name(f'{self.name}_{parameter}_histogram').with_suffix('.png'))

    return figure, axis



def histogram_FRET(data, axis, **kwargs):
    axis.hist(data, range=(0, 1), **kwargs)
    axis.set_xlim((0, 1))
    axis.set_xlabel('FRET')
    axis.set_ylabel('Count')

def fit_hist(data, axis):
    hist, bin_edges = np.histogram(data, 100, range=(0, 1))
    bin_centers = (bin_edges[0:-1] + bin_edges[1:]) / 2

    # plt.plot(bin_centers,hist)

    from scipy.signal import butter
    from scipy.signal import filtfilt
    b, a = butter(2, 0.2, 'low')
    output_signal = filtfilt(b, a, hist)
    plt.plot(bin_centers, output_signal)

    from scipy.signal import find_peaks
    peaks, properties = find_peaks(output_signal, prominence=5, width=7)  # prominence=1
    plt.plot(bin_centers[peaks], hist[peaks], "x")

    def func(x, a, b, c, d, e, f):
        return a * np.exp(-(x - b) ** 2 / (2 * c ** 2)) + d * np.exp(-(x - e) ** 2 / (2 * f ** 2))

    from scipy.optimize import curve_fit
    popt, pcov = curve_fit(func, bin_centers, hist, method='trf',
                           p0=[hist[peaks[0]], bin_centers[peaks[0]], 0.1, hist[peaks[1]], bin_centers[peaks[1]], 0.1],
                           bounds=(0, [np.inf, 1, 1, np.inf, 1, 1]))

    axis.plot(bin_centers, func(bin_centers, *popt))
    # plt.plot(bin_centers,func(bin_centers, 10000,0.18,0.1,5000,0.5,0.2))

# uniqueFileNames = list(set([re.search('hel[0-9]*',fileName).group() for fileName in fileNames]))


def show_image_3d(image, figure=None):
    if not figure:
        figure = plt.figure()

    from matplotlib import cm
    axis = figure.gca(projection='3d')
    X = np.arange(image.shape[1])
    Y = np.arange(image.shape[0])
    X, Y = np.meshgrid(X, Y)
    axis.plot_surface(X, Y, image, cmap=cm.coolwarm,
                      linewidth=0, antialiased=False)




# -*- coding: utf-8 -*-
"""
Created on Mon Mar  8 16:10:42 2021

input = image, coordinates, method, filter_neighbourhood_size, radius
methods+args:
    Channel_mean
    Channel_median
    ROI_minimum, uses filter_neighbourhood_size ** DEFAULT
    ROI_median, uses filter_neighbourhood_size
    Surrounding_mean, uses radius
    Surrounding_median, uses radius
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.ndimage.filters as filters


def extract_background(image, coordinates, method='ROI_minimum', filter_neighbourhood_size=10, radius=5):
    background = np.zeros(coordinates.shape[0])

    if method == 'channel_mean':
        background[:] = np.mean(image)

    elif method == 'channel_median':
        background[:] = np.median(image)

    elif method == 'ROI_minimum':
        image_min = filters.minimum_filter(image, filter_neighbourhood_size)
        # TODO: improve edge between donor&acceptor channel
        # plt.imshow(image_min)
        for i, (x, y) in enumerate(coordinates):
            background[i] = image_min[int(y), int(x)]

    elif method == 'ROI_median':
        image_median = filters.median_filter(image, filter_neighbourhood_size)
        # TODO: improve edge between donor&acceptor channel
        for i, (x, y) in enumerate(coordinates):
            background[i] = image_median[int(y), int(x)]

    elif method == 'surrounding_mean':  # not optimal when multiple spots are close, better to use median
        from papylio.coordinate_optimization import crop, circle
        circle_matrix = circle(radius)
        for i, c in coordinates:
            cropped_peak = crop(image, c, radius * 2 + 1)
            circle_image = cropped_peak * circle_matrix
            background[i] = np.mean(circle_image[np.nonzero(circle_image)])

    elif method == 'surrounding_median':  # similar to coordinate_optimization, coordinates_without_intensity_at_radius
        from papylio.coordinate_optimization import crop, circle
        circle_matrix = circle(radius)
        for i, c in coordinates:
            cropped_peak = crop(image, c, radius * 2 + 1)
            circle_image = cropped_peak * circle_matrix
            background[i] = np.median(circle_image[np.nonzero(circle_image)])

    else:
        raise ValueError('Unknown method')

    return background



import pytest
import tifffile
import numpy as np


@pytest.fixture
def experiment(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF')

@pytest.fixture
def file(experiment):
    return experiment.files[1]

@pytest.fixture
def experiment_output(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF_output_test_file')

@pytest.fixture
def file_output(experiment_output):
    return experiment_output.files[1]

@pytest.fixture
def file_output_with_selected(file_output):
    #TODO: This is not optimal, perhaps add selection to the file.
    selected = file_output.selected
    selected[[0,5,33]] = True
    file_output.set_variable(selected)
    return file_output

def test_projection_image(file, shared_datadir):
    image_newly_made = file.projection_image()
    image_from_original_file = tifffile.imread(shared_datadir / 'BN_TIRF_output_test_file' / 'TIRF 561 0001_ave_f0-10_i0.tif')
    assert (image_newly_made == image_from_original_file).all()
    image_loaded = file.projection_image()
    assert (image_loaded == image_from_original_file).all()

def test_average_image(file, shared_datadir):
    image_newly_made = file.average_image()
    image_from_original_file = tifffile.imread(shared_datadir / 'BN_TIRF_output_test_file' / 'TIRF 561 0001_ave_f0-10_i0.tif')
    assert (image_newly_made == image_from_original_file).all()
    image_loaded = file.average_image()
    assert (image_loaded == image_from_original_file).all()

def test_maximum_projection_image(file, shared_datadir):
    image_newly_made = file.maximum_projection_image()
    image_from_original_file = tifffile.imread(shared_datadir / 'BN_TIRF_output_test_file' / 'TIRF 561 0001_max_f0-10_i0.tif')
    assert (image_newly_made == image_from_original_file).all()
    image_loaded = file.maximum_projection_image()
    assert (image_loaded == image_from_original_file).all()

def test_perform_mapping(experiment, shared_datadir):
    import matchpoint as mp
    experiment.files[0].perform_mapping()
    mapping_control = mp.MatchPoint.load(shared_datadir / 'BN_TIRF_output_test_file' / 'beads.mapping')
    assert ((experiment.files[0].mapping.transformation.params - mapping_control.transformation.params) < 1e-2).all()

def test_find_molecules(file):
    file.find_coordinates()

def test_extract_traces(file):
    file.find_coordinates()
    file.extract_traces()
    file.extract_traces(mask_size=None, neighbourhood_size=None,
                        background_correction=(-150,-30),
                        alpha_correction=0.075,
                        gamma_correction=1.2)

def test_property_coordinates(file_output):
    file_output.coordinates

def test_determine_psf_size(file):
    psf_size = file.determine_psf_size()
    assert np.isclose(psf_size, 1.01, atol=0.005)

def test_show_histogram(file_output):
    file_output.show_histogram('intensity')
    file_output.show_histogram('FRET', bins=100, range=(0,1))

def test_show_traces(file_output):
    file_output.show_traces(selected=False)

def test_save_dataset_selected(file_output_with_selected):
    file_output_with_selected.save_dataset_selected()
    import xarray as xr
    ds = xr.load_dataset(file_output_with_selected.absoluteFilePath.parent / (file_output_with_selected.name + '_selected.nc'))
    indices_selected = np.nonzero(ds.molecule_in_file.values)[0]
    assert (indices_selected == np.array([0,5,33])).all().item()

def test_classify_hmm(file_output):
    selection = file_output.selected
    selection[0:20] = True
    file_output.set_variable(selection, name='selected')
    file_output.apply_classifications()
    file_output.classify_hmm('FRET')
    file_output.classify_hmm(file_output.intensity.sel(channel=0, drop=True))

def test_use_for_darkfield_correction(file):
    file.use_for_darkfield_correction()


import os
import numpy as np
import xarray as xr
import sys
import pandas as pd
import matplotlib.pyplot as plt
import scipy.optimize
import inspect
import seaborn as sns

def dwell_frames_from_classification(classification):
    # This assumes continuous monitoring with a specific cycle time.
    single_true_array = np.ones((classification.shape[0],1)).astype(bool)
    is_state_transition = np.hstack([single_true_array, classification[:,:-1] != classification[:,1:], single_true_array])
    state_transition_molecule, state_transition_frame = np.where(is_state_transition)
    is_endpoint = state_transition_frame == classification.shape[1]
    dwell_states = classification[state_transition_molecule[~is_endpoint], state_transition_frame[~is_endpoint]]
    dwell_frames = np.diff(state_transition_frame)[~is_endpoint[:-1]]
    dwell_molecules = state_transition_molecule[~is_endpoint]
    return dwell_molecules, dwell_states, dwell_frames

def determine_dwell_means(traces_flattened, dwell_frames):
    mean_trace = np.mean(traces_flattened)
    values_cumsum = np.concatenate([[0], np.cumsum(traces_flattened-mean_trace)])
    oneD_indices = np.concatenate([[0],dwell_frames.cumsum()])
    dwell_means = np.diff(values_cumsum[oneD_indices]) / dwell_frames + mean_trace
    return dwell_means

def set_states(dwell_molecules, dwell_states, at_trace_edges=True, around_negative_states=True, to_state=-2):
    states_to_set = np.zeros(len(dwell_states), dtype=bool)

    switched_molecule = np.diff(dwell_molecules).astype(bool)
    start_and_end_trace = np.concatenate([[True], switched_molecule]) | np.concatenate([switched_molecule, [True]])

    negative_states = dwell_states < 0

    if at_trace_edges:
        states_to_set |= start_and_end_trace

    if around_negative_states:
        negative_state_neighbours = np.concatenate([[False], negative_states[:-1]]) | \
                                    np.concatenate([negative_states[1:], [False]])
        states_to_set |= negative_state_neighbours & ~start_and_end_trace

    states_to_set[negative_states] = False

    dwell_states[states_to_set] = to_state
    return dwell_states


def dwell_times_from_classification(classification, traces=None, cycle_time=None, inactivate_start_and_end_states=True):
    if isinstance(classification, xr.DataArray):
        molecule_coords = {n: c.values for n, c in classification.coords.items() if c.dims[0] == 'molecule' and len(c.dims) == 1}
        classification = classification.values
    else:
        molecule_coords = None
    dwell_molecules, dwell_states, dwell_frames = dwell_frames_from_classification(classification)
    if inactivate_start_and_end_states:
        dwell_states = set_states(dwell_molecules, dwell_states, to_state=-128)
        # Probably better to indicate for each dwell whether it is at a trace edge or around a negative state

    ds = xr.Dataset()
    if molecule_coords is not None:
        for n, c in molecule_coords.items():
            ds[n] = ('dwell', c[dwell_molecules])
    else:
        ds['molecule'] = ('dwell', dwell_molecules)

    ds['state'] = ('dwell', dwell_states)
    ds['frame_count'] = ('dwell', dwell_frames)

    if cycle_time is not None:
        dwell_times = dwell_frames * cycle_time
        ds['duration'] =('dwell', dwell_times)

    if traces is not None:
        # if isinstance(traces, xr.Dataset):
        #     for name, da in traces.data_vars.items():
        #         dwell_means = determine_dwell_means(da.values.flatten(), dwell_frames)
        #         ds['mean'] = xr.DataArray(dwell_means, dims=['dwell'])
        name = ''
        if isinstance(traces, xr.DataArray):
            if traces.name is not None:
                name = '_' + traces.name
            traces = traces.values

        dwell_means = determine_dwell_means(traces.flatten(), dwell_frames)
        ds['mean'+name] = ('dwell', dwell_means)

    return ds


# def single_decaying_exponential(t, A, tau):
#     return A * np.exp(-t/tau)
# single_decaying_exponential.bounds = ((0,0),(np.inf,np.inf))
# def p0(t, y):
#     return y.max(), t.mean()
# single_decaying_exponential.p0 = p0

def single_decaying_exponential(t, A, k):
    return A * np.exp(-k*t)
single_decaying_exponential.bounds = ((0,0),(np.inf,np.inf))
def p0(t, y):
    return y.max(), 1/t.mean()
single_decaying_exponential.p0 = p0

def analyze_dwells(dwells, fit_function=single_decaying_exponential, cycle_time=1, plot=False,
                   axes=None, state_names={0: 'Low FRET state', 1: 'High FRET state'}, logy=False, sharey=True):
    # states = np.unique(dwells.state)
    states = np.array(list(state_names.keys()))
    positive_states = states[states>=0]

    bins=50

    if plot and axes is None:
        fig, axes = plt.subplots(1,len(positive_states), figsize=(len(positive_states)*3, 2), layout='constrained', sharey=sharey)
    else:
        axes = None

    fit_parameters = list(inspect.signature(fit_function).parameters)[1:]
    fit_values = xr.Dataset(coords={'state': positive_states, 'parameter': fit_parameters})
    fit_values['optimal_value'] = xr.DataArray(np.nan, dims=('state', 'parameter'), coords={'state': positive_states, 'parameter': fit_parameters})
    fit_values['error'] = xr.DataArray(np.nan, dims=('state', 'parameter'), coords={'state': positive_states, 'parameter': fit_parameters})
    fit_values['covariance'] = xr.DataArray(np.nan, dims=('state', 'parameter','parameter'),
                                       coords={'state': positive_states, 'parameter': fit_parameters, 'parameter': fit_parameters})

    fit_values.attrs['fit_function'] = fit_function.__name__

    for i, state in enumerate(positive_states):
        dwells_with_state = dwells.sel(dwell=dwells.state==state)
        c, t_edges = np.histogram(dwells_with_state.duration, bins=bins+1, range=[-cycle_time/2, (bins+1/2)*cycle_time])
        t = (t_edges[:-1]+t_edges[1:])/2
        try:
            popt, pcov = scipy.optimize.curve_fit(fit_function, t[2:], c[2:], p0=fit_function.p0(t[2:], c[2:]),
                                                  bounds=fit_function.bounds, absolute_sigma=True)
            perr = np.sqrt(np.diag(pcov))
            fit_values['optimal_value'][dict(state=state)] = popt
            fit_values['covariance'][dict(state=state)] = pcov
            fit_values['error'][dict(state=state)] = perr

            # fit_values[state] = {fit_parameter: {'value': value, 'error': error} for fit_parameter, value, error in
            #                      zip(fit_parameters, popt, perr)}

        except RuntimeError:
            popt = pcov = perr = None

        if plot:
            # TODO: make this a separate function
            axes[i].bar(t, c, width=cycle_time)
            # axes[i].set_title(+ ',' + str())
            if popt is not None:
                axes[i].plot(t, fit_function(t, *popt), c='r', label=r'$count = Ae^{-\frac{t}{\tau}}$')
                text_string = f'A={popt[0]:.5}\nτ={popt[-1]:.5}\n1/τ={1/popt[-1]:.5}' #\nmean={dwells_with_state["mean"].mean().item():.5}'
                text_string = f'A={popt[0]:.0f}±{perr[0]:.0f}\nk={popt[-1]:.2f}±{perr[-1]:.2f}\nτ={1 / popt[-1]:.1f}'  # \nmean={dwells_with_state["mean"].mean().item():.5}'
            else:
                text_string = 'No fit found'

            text_string = state_names[state] + '\n' + text_string

            # axes[i].text(0.95, 0.95, text_string, transform=axes[i].transAxes, fontsize=14,
            #     verticalalignment='top')# , bbox=props)
            axes[i].annotate(text_string, (0.55, 0.9), xycoords='axes fraction', fontsize=9,
                        verticalalignment='top')


            if axes[i].get_subplotspec().is_last_row():
                axes[i].set_xlabel('Dwell time (s)')
            axes[i].set_xlim([-cycle_time/2,30*cycle_time])
            if axes[i].get_subplotspec().is_first_col():
                # axes[i].set_ylim(0,c[1:].max()*1.1)
                axes[i].set_ylabel('Dwell count')
                if logy:
                    axes[i].set_yscale('log')
                    axes[i].set_ylim(0.5, axes[i].get_ylim()[1])
            # if axes[i].get_subplotspec().is_last_col():
            #     axes[i].legend()

    return fit_values, axes



if __name__ == '__main__':
    classification = np.array([-1,-1,2,2,2,2,1,1,2,2,2,1,1,1,-1,-1,0,0,0,0,0,0,0,0,0,0,2,2,0,0,0,0,0])
    classification = np.repeat(classification[None,:], 5, axis=0)
    traces = np.random.random(classification.shape)

    dwell_times_from_classification(classification, traces=traces, cycle_time=0.1)

#
#
# classification = np.array([-1,-1,2,2,2,2,1,1,2,2,2,1,1,1,2,2,0,0,0,0,0,0,0,0,0,0,-1,-1,0,0,0,0,0])
# trace = np.random.random(classification.shape)
#
#
# def dwell_times_from_classification(classification, cycle_time=1):
#     # This assumes continuous monitoring with a specific cycle time.
#     is_state_transition = np.concatenate([[True], classification[:-1] != classification[1:], [True]])
#     state_transition_indices = np.where(is_state_transition)[0]
#     dwell_states = classification[state_transition_indices[:-1]]
#     dwell_frames = np.diff(state_transition_indices)
#     dwell_times = dwell_frames * cycle_time
#
#     mean_trace = np.mean(trace)
#     values_cumsum = np.concatenate([[0], np.cumsum(trace-mean_trace)])
#     dwell_means = np.diff(values_cumsum[state_transition_indices]) / dwell_frames + mean_trace
#
#     return dwell_states, dwell_times, dwell_means
#
#
# dwell_times_from_classification(classification)














# if __name__ == '__main__':
#     import SAfitting
#     import common_PDF
# else:
#     from papylio.analysis import SAfitting
#     from papylio.analysis import common_PDF
# # import SAfitting
# sns.set(style="ticks")
# sns.set_color_codes()
#
#
# def analyze(dwells_data, dataset_name, dist, configuration):
#     conf = configuration
#     # find the Tmax until which data is selected
#     d = apply_config_to_data(dwells_data, dist, conf)
#     figures = []
#     fit_data = []
#     keys_with_data = []  # keys refer to 'red', 'green', 'total', 'FRET'
#     for key in d.keys():
#         if d[key].empty:  # check if the dataframe is empty
#             print(f'{dist} dataFrame for {key} is empty')
#             continue
#         dwells = d[key].loc[:,dist].values
#         dwells = dwells[dwells>0]
#         print(np.size(dwells), 'dwells selected')
#         if conf['FitBool']:
#             fit_res = fit(dwells, model=conf['model'], dataset_name=dataset_name,
#                           Nfits=int(conf['Nfits']),
#                           include_over_Tmax=conf['TmaxBool'],
#                           bootstrap=conf['BootBool'],
#                           boot_repeats=int(conf['BootRepeats']))
#             fit_data.append(fit_res)
#         else:
#             fit_res = None
#         print(f'plotting {key} {dist}')
#         figure = plot(dwells, dataset_name, dist, trace=key, binsize=conf['binsize'],
#                       scale=conf['scale'], style=conf['PlotType'],
#                       fit_result=fit_res)
#         figures.append(figure)
#         keys_with_data.append(key)
#
#     if fit_data != []:
#         fit_data = pd.concat(fit_data, axis=1, keys=keys_with_data)
#     return d, figures, fit_data
#
# def fit(dwells, model='1Exp', dataset_name='Dwells', Nfits=1,
#         include_over_Tmax=True, bootstrap=True, boot_repeats=100):
#
#     if model == '1Exp+2Exp':
#         fit_result = []
#         for model in ['1Exp', '2Exp']:
#             result, boots = SAfitting.fit(dwells, model, dataset_name, Nfits,
#                                    include_over_Tmax, bootstrap, boot_repeats)
#             fit_result.append(result)
#         fit_result = pd.concat(fit_result, axis=1, ignore_index=True)
#         return fit_result
#
#     fit_result, boots = SAfitting.fit(dwells, model, dataset_name, Nfits, include_over_Tmax,
#                                   bootstrap, boot_repeats)
#     # print(fit_result)
#     return fit_result
#
#
# def plot(dwells, name, dist='offtime', trace='red', binsize='auto', scale='log',
#          style='dots', color='from_trace', fit_result=None):
#
#     if fit_result is not None:
#         if fit_result.Ncut[0] > 0:
#             Tcut = dwells.max() - 5  # 5 sec is kind of arbitrary here
#             dwells = dwells[dwells < Tcut]
#
#     try:
#         bsize = float(binsize)
#         if scale == 'Log-Log':
#             bin_edges = 10**(np.arange(np.log10(min(dwells)), np.log10(max(dwells)) + bsize, bsize))
#         else:
#             bin_edges = np.arange(min(dwells), max(dwells) + bsize, bsize)
#     except ValueError:
#         if binsize == 'Auto':
#             binsize = 'auto'
#         bin_edges = binsize
#     values, bins = np.histogram(dwells, bins=bin_edges, density=True)
#
#     # Determine position of bins
#     if scale == 'Log-Log':
#         centers = (bins[1:] * bins[:-1])**0.5  # geometric average of bin edges
#     else:
#         centers = (bins[1:] + bins[:-1]) / 2.0
#
#     # combine bins until they contain at least one data point (for y-log plots)
#     if scale in ['Log', 'Log-Log']:
#         izeros = np.where(values == 0)[0]
#         print('izeros', izeros)
#         j = 0
#         while j < len(izeros):
#             i = j
#             j += 1
#             while j < len(izeros) and izeros[j] - izeros[j-1] == 1:
#                 j += 1
#             # print('jstart ', izeros[i])
#             # print('jend ', izeros[i]+(j-i))
#             # print('values ', values[izeros[i]:(izeros[i]+j-i+1)])
#             # print('mean value', np.sum(values[izeros[i]:(izeros[i]+j-i+1)])/(j-i+1))
#             values[izeros[i]:(izeros[i]+j-i+1)] = np.sum(values[izeros[i]:(izeros[i]+j-i+1)])/(j-i+1)
#
#     fig = plt.figure(f'Histogram {trace} {dist}s {name}', figsize=(4, 3), dpi=200)
#
#     if color == 'from_trace':
#         if dist == 'offtime':
#             color = 'r'*(trace == 'red') + 'g'*(trace == 'green') + \
#                     'b'*(trace == 'FRET') + 'sandybrown'*(trace == 'total')
#         if dist == 'ontime':
#             color = 'firebrick'*(trace == 'red') + 'olive'*(trace == 'green') + \
#                     'darkviolet'*(trace == ' FRET') + 'saddlebrown'*(trace == 'total')
#     label = f'{dist} pdf, N={dwells.size}'
#     if style == 'dots':
#         plt.plot(centers, values, '.', color=color, label=label)
#     if style == 'bars':
#         plt.bar(centers, values, color=color, label=label,
#                 width=(bins[1] - bins[0]))
#     if style == 'line':
#         plt.plot(centers, values, '-', lw=2, color=color, label=label)
#
#     if fit_result is not None:
#         if fit_result.model[0] == '1Exp':
#             tau = fit_result.value[0]
#             error = fit_result.error[0]
#             Ncut = fit_result.Ncut[0]
#             print(f'plotting 1Exp fit')
#             time, fit = common_PDF.Exp1(tau,
#                                         Tmax=centers[-1]+(bins[1]-bins[0])/2)
#             label = f'\n tau={tau:.1f}'
#             if error != 0:
#                 label += f'$\pm$ {error:.1f}'
# #            plt.plot(time, fit, color='r', label=f'1expFit, Ncut={int(Ncut)} \n {label}')
#
#         elif fit_result.model[0] == '2Exp':
#             p, errp = fit_result.value[0], fit_result.error[0]
#             tau1, err1 = fit_result.value[1], fit_result.error[1]
#             tau2, err2 = fit_result.value[2], fit_result.error[2]
#             Ncut = fit_result.Ncut[0]
#             print(fit_result)
#             print(f'errors: ', errp, err1, err2)
#             time, fit = common_PDF.Exp2(p, tau1, tau2, Tmax=centers[-1])
#             label = f'\n p={p:.2f}, tau1={tau1:.1f}, tau2={int(tau2)}'
# #            plt.plot(time, fit, color='r', label=f'2expFit, Ncut={int(Ncut)} \n {label}')
#
#         elif fit_result.model[0] == '3Exp':
#             p1, errp1 = fit_result.value[0], fit_result.error[0]
#             p2, errp2 = fit_result.value[1], fit_result.error[1]
#             tau1, err1 = fit_result.value[2], fit_result.error[2]
#             tau2, err2 = fit_result.value[3], fit_result.error[3]
#             tau3, err3 = fit_result.value[4], fit_result.error[4]
#             Ncut = fit_result.Ncut[0]
#             print(fit_result)
#             print(f'errors: ', errp1, errp2, err1, err2, err3)
#             time, fit = common_PDF.Exp3(p1, p2, tau1, tau2, tau3,
#                                         Tmax=centers[-1])
#             label = f'\n p1={p1:.2f}, p2={p2:.2f}, tau1={tau1:.1f}, tau2={int(tau2)}, tau3={int(tau3)}'
# #            plt.plot(time, fit, color='r', label=f'3expFit, Ncut={int(Ncut)} \n {label}')
#
#         if fit_result.Ncut[0] > 0:
#             label = f', Ncut={int(Ncut)}' + label
#
#         plt.plot(time, fit, color='k', label=f'{fit_result.model[0]}fit{label}')
#
#     if scale in ['Log', 'Log-Log']:
#         plt.yscale('log')
#
#     if scale == 'Log-Log':
#         plt.xscale('log')
#
#     plt.legend()
#     plt.ylabel('Probability')
#     plt.xlabel(f'{dist} (s)')
#     # plt.locator_params(axis='y', nbins=3)
#     plt.tight_layout()
#     plt.show()
#     return fig
#
# def apply_config_to_data(dwells_data, dist, config):
#     d = dwells_data
#     # Select the requested sides
#     side_list = ['l'*bool(config['side']['left']),
#                'm'*bool(config['side']['middle']),
#                'r'*bool(config['side']['right'])]
#
#     if dist == 'offtime':
#         d = d[d.side.isin(side_list)]
#     if dist == 'ontime':
#         d = d[d.onside.isin(side_list)]
#     # apply min, max conditions
#     if config['max'] in ['Max', 'max']:
#         d = d[d[dist] > float(config['min'])]
#     else:
#         d = d[d[dist] > float(config['min'])]
#         d = d[d[dist] < float(config['max'])]
#
#     data = {}
#
#     for key in config['trace'].keys():
#         if config['trace'][key]:
#             data[key] = d[d['trace'] == key]
#         else:
#             pass
#
#     return data
#
#
# if __name__ == '__main__':
#     filename = 'C:/Users/iason/Desktop/traceanalysis/papylio/traces/'
#     filename += 'hel0_dwells_data.xlsx'
#
#     data = pd.read_excel(filename, index_col=[0, 1], dtype={'kon' :np.str})
#     print(data.shape)
#     config = {'trace': {'red': True, 'green': False, 'total': False, 'FRET': False},
#          'side': {'left': True, 'middle': True, 'right': True},
#          'min': '0', 'max': 'max',
#          'scale': 'Normal',
#          'PlotType': 'dots',
#          'binsize': 'auto',
#          'FitBool': True,
#          'TmaxBool': False,
#          'BootBool': False,
#          'model': '2Exp',
#          'Nfits': '1',
#          'BootRepeats': '5'}
#
#     result = analyze(data, 'test', 'offtime', config)



import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.path as pth
import math
from pathlib import Path
from skimage.transform import AffineTransform, SimilarityTransform
import tqdm
import pandas as pd
import xarray as xr
import os.path
# from papylio.experiment import Experiment
# from papylio.file import File
# from papylio.mapping.geometricHashing import SequencingDataMapping
import matchpoint as mp
from papylio.decorators import return_none_when_executed_by_pycharm
from .fastqAnalysis import FastqData
from .geometricHashing2 import geometric_hash, find_match_after_hashing
from .geometricHashing3 import GeometricHashTable
from .plotting import plot_sequencing_match, plot_matched_files_in_tile
from .sequencing_data import SequencingData, make_sequencing_dataset
from .mapping_collection import MappingCollection



class Experiment:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sequencing_data_for_mapping = None # To be removed
        self._tile_mappings = None

        if 'sequencing' in self.configuration.keys():
            sequencing_data_relative_file_path = self.configuration['sequencing']['data_file_path']
            sequencing_data_file_path = self.main_path.joinpath(sequencing_data_relative_file_path)
            # This should work fine for the case where it is a relative or absolute path
            if sequencing_data_file_path is not None:
                self.sequencing_data = SequencingData(sequencing_data_file_path, load=False)
                print(f"\nImport sequencing data:\n{sequencing_data_relative_file_path}")

    @property
    @return_none_when_executed_by_pycharm
    def tile_mappings(self):
        if self._tile_mappings is None:
            tile_mappings_path = self.main_path.joinpath('Analysis').joinpath('Tile mappings')
            self._tile_mappings = MappingCollection([mp.MatchPoint.load(filepath) for filepath in
                                                     tile_mappings_path.glob('*.nc')])

        return self._tile_mappings

    @property
    @return_none_when_executed_by_pycharm
    def tile_mappings_dict(self):
        d = {}
        for mapping in self.tile_mappings:
            name = mapping.name[:-12]
            if name not in d.keys():
                d[name] = {}
            d[name][mapping.label] = mapping
        return d

    def import_sequencing_data(self, file_path, index1_file_path=None, remove_duplicates=True,
                               add_aligned_sequence=True, extract_sequence_subset=False, chunksize=10000,
                               store_relative_filepath=True):
        file_path = Path(file_path)
        if file_path.suffix == '.csv':
            raise ValueError('Wrong file type for sequencing data, if you would like to import the old .csv files, use "import_sequencing_data_old" ')

        nc_file_path = make_sequencing_dataset(file_path, index1_file_path=index1_file_path,
                                               remove_duplicates=remove_duplicates,
                                               add_aligned_sequence=add_aligned_sequence,
                                               extract_sequence_subset=extract_sequence_subset, chunksize=chunksize)
        self.sequencing_data = SequencingData(nc_file_path, load=False)
        # TODO: Insert full path if the path is on another drive.
        if store_relative_filepath:
            nc_file_path = os.path.relpath(nc_file_path, start=self.main_path)
        self.configuration['sequencing'] = {'data_file_path': nc_file_path}
        self.configuration.save()

    def import_sequencing_data_old(self):
        raise DeprecationWarning('import_sequencing_data_old will be removed')
        seqdata = SequencingData(file_path)
        if surface == 0:
            seqdata = seqdata[seqdata.tile < 2000]
        elif surface == 1:
            seqdata = seqdata[seqdata.tile > 2000]
        else:
            raise ValueError('Surface can be either 0 or 1')

        self.sequencing_data = seqdata

    def import_sequencing_data_for_mapping(self, file_path, surface=0):
        #TODO: Merge with import_sequencing_data to obtain single
        raise DeprecationWarning('import_sequencing_data_for_mapping will be removed')
        seqdata = SequencingData(file_path)
        if surface == 0:
            seqdata = seqdata[seqdata.tile < 2000]
        elif surface == 1:
            seqdata = seqdata[seqdata.tile > 2000]
        else:
            raise ValueError('Surface can be either 0 or 1')

        self.sequencing_data_for_mapping = seqdata

    def generate_tile_mappings(self, files_for_mapping, mapping_sequence_name=None, surface=0, name='All files'):
        coordinates_sm = files_for_mapping.coordinates_stage
        if self.sequencing_data_for_mapping is not None and mapping_sequence_name is None:
            coordinates_seq = self.sequencing_data_for_mapping.coordinates ## To be removed
        else:
            selection = (self.sequencing_data.dataset.reference_name == mapping_sequence_name)
            if surface == 0:
                selection &= self.sequencing_data.dataset.tile < 2000
            elif surface == 1:
                selection &= self.sequencing_data.dataset.tile > 2000
            else:
                raise ValueError('Surface can be either 0 or 1')
            sequencing_data = self.sequencing_data[selection]
            coordinates_seq = sequencing_data.coordinates
            tiles_seq = sequencing_data.tile

        tile_mapping_path = self.analysis_path.joinpath('Tile mappings')
        tile_mapping_path.mkdir(parents=True, exist_ok=True)

        tile_mappings = []
        for tile in tqdm.tqdm(sequencing_data.tile_numbers, 'Make tile mappings'):
            mapping = mp.MatchPoint(source=coordinates_sm, destination=coordinates_seq.sel(sequence=tiles_seq==tile))
            mapping.transformation_type = 'linear'
            mapping.name = f'Tile {tile}'
            mapping.label = tile
            mapping.tile = tile # Is this one still necessary?
            mapping.source_name = 'Single-molecule data'
            mapping.destination_name = 'Sequencing data'
            mapping.source_unit = 'µm'
            mapping.destination_unit = 'MiSeq'
            mapping.save(tile_mapping_path.joinpath(f'{name} - Tile {tile}.nc'))
            tile_mappings.append(mapping)
        self._tile_mappings = None

    def transform_sequencing_to_single_molecule_coordinates(self):
        coordinates_seq = self.sequencing_data.coordinates
        coords = coordinates_seq.coords
        coordinates_sm = xr.DataArray(dims=['mapping_name','sequence','dimension'],
                                      coords={'mapping_name': list(self.tile_mappings_dict.keys()), 'sequence': coords['sequence'],
                                              'dimension': coords['dimension'], 'tile': coords['tile']})
        tile_mappings_dict = self.tile_mappings_dict
        # for tile, coordinates in tqdm.tqdm(self.sequencing_data.coordinates.groupby('tile'), unit='tile'):
        for tile in tqdm.tqdm(self.sequencing_data.tile_numbers, unit='tile'):
            for mapping_name, tile_mappings in tile_mappings_dict.items():
                if tile in tile_mappings.keys():
                    coordinates_sm.loc[dict(mapping_name=mapping_name, sequence=(coordinates_sm.tile == tile))] = \
                       tile_mappings[tile].transformation_inverse(coordinates_seq.sel(sequence=coordinates_seq.tile==tile))
                # coordinates_sm.append(mapping.transform_coordinates(coordinates, inverse=True))

        self.sequencing_data.dataset['x_sm'] = coordinates_sm.sel(dimension='x')
        self.sequencing_data.dataset['y_sm'] = coordinates_sm.sel(dimension='y')

        # self.sequencing_data.dataset.plot.scatter('y_sm', 'x_sm')















    # def import_sequencing_data(self, fastq_file_path):
    #     self.fastq_file_path = Path(fastq_file_path)
    #     self.sequencing_data = FastqData(self.fastq_file_path)

    # def generate_mapping_hashtable(self, imaged_surface=None, maximum_distance_tile=None, tuple_size=None):
    #
    #     # TODO: Add timer to generate_mapping_hashtable and find_sequences methods, by making a decorator function. [IS: 10-08-2020]
    #
    #     # self.select_sequencing_data_for_mapping(mapping_sequence, number_of_allowed_mismatches)
    #
    #     if imaged_surface in ['top', 1]:
    #         self.sequencing_data_for_mapping = self.sequencing_data_for_mapping[self.sequencing_data_for_mapping.tile < 2000]
    #     elif imaged_surface in ['bottom', 2]:
    #         self.sequencing_data_for_mapping = self.sequencing_data_for_mapping[self.sequencing_data_for_mapping.tile > 2000]
    #
    #     tile_coordinate_sets = [tile.coordinates for tile in self.sequencing_data_for_mapping.tiles]
    #     # TODO: get maximum_distance_tile and tuple_size from configuration
    #     self.geometric_hash_data = geometric_hash(tile_coordinate_sets, maximum_distance_tile, tuple_size)

    # def generate_mapping_hashtable3(self, imaged_surface=None, initial_file_transformation=None, maximum_distance_tile=None, tuple_size=None):
    #
    #     # TODO: Add timer to generate_mapping_hashtable and find_sequences methods, by making a decorator function. [IS: 10-08-2020]
    #
    #     # self.select_sequencing_data_for_mapping(mapping_sequence, number_of_allowed_mismatches)
    #
    #     if imaged_surface in ['top', 1]:
    #         sequencing_data_for_mapping = self.sequencing_data_for_mapping[self.sequencing_data_for_mapping.tile < 2000]
    #     elif imaged_surface in ['bottom', 2]:
    #         sequencing_data_for_mapping = self.sequencing_data_for_mapping[self.sequencing_data_for_mapping.tile > 2000]
    #
    #     tile_coordinate_sets = [tile.coordinates for tile in sequencing_data_for_mapping.tiles]
    #
    #     # initial_magnification = np.array([3.67058194, -3.67058194])
    #     # initial_rotation = 0.6285672733195177  # degrees
    #     # initial_file_transformation = AffineTransform(matrix=None, scale=initial_magnification,
    #     #                                                 rotation=initial_rotation/360*np.pi*2,
    #     #                                                 shear=None, translation=None)
    #     self.geometric_hashtable = GeometricHashTable(tile_coordinate_sets,
    #                                                   source_vertices=self.files[0].movie.channels[1].vertices,
    #                                                   initial_source_transformation=initial_file_transformation)
    #
    # def generate_mapping_hashtable_from_coordinate_set(self, tile_coordinate_sets, maximum_distance_tile, tuple_size):
    #     self.geometric_hash_data = geometric_hash(tile_coordinate_sets, maximum_distance_tile, tuple_size)
    #
    # def select_sequencing_data_for_mapping(self, mapping_sequence, number_of_allowed_mismatches):
    #     self.mapping_sequence = mapping_sequence
    #     self.sequencing_data.matches_per_tile(sequence=mapping_sequence)
    #
    #     number_of_matches = self.sequencing_data.number_of_matches(mapping_sequence)
    #     # sequencing_data_for_mapping = sequencing_data.select(Nmatch == len(mapping_sequence), copyData=True)
    #     self.sequencing_data_for_mapping = self.sequencing_data[number_of_matches >=
    #                                                             (len(mapping_sequence) - number_of_allowed_mismatches)]
    #     self.sequencing_data_for_mapping.show_tiles()
    #     self.sequencing_data_for_mapping.export_positions_per_tile()

    # def map_files_to_sequencing_data(self, tile, configuration=None):
    #     sequencing_mapping_path = self.main_path.joinpath('sequencing_mapping')
    #     sequencing_mapping_path.mkdir(exist_ok=True)
    #
    #     presets = {'TIR-I single-molecule':
    #                    {
    #                        'initial_image_transformation': {'reflection': 0, 'rotation': np.pi, 'magnification': 3.336},
    #                        'mapping_configuration': {'mode': 'similarity',
    #                                                 'nBins': 250,
    #                                                 'rotationRange': np.array([-180, 180]) / 360 * 2 * np.pi,
    #                                                 'magnificationRange': np.array([3000, 6000])
    #                                                 },
    #                        'bases_findMatch': 500
    #                     }
    #                }
    #     # rotationRange = np.array([-5,5])/360*2*np.pi+np.pi/2,
    #     # magnificationRange = np.array([2750,3000])
    #     # seqmap.initial_image_transformation = {'reflection': 0, 'rotation': -35/180*np.pi, 'magnification':1/125 }
    #     # seqmap.initial_image_transformation = {'reflection': 0, 'rotation': np.pi, 'magnification': 2*3.336}
    #     # seqmap.initial_image_transformation = {}
    #
    #     if type(configuration) is str:
    #         configuration = presets[configuration]
    #
    #     tile = self.sequencing_data_for_mapping.get_tile_object(tile=tile)
    #     self.seqmap = SequencingDataMapping(tile, self.files, sequencing_mapping_path, **configuration['mapping_configuration'])
    #     self.seqmap.initial_image_transformation = configuration['initial_image_transformation']
    #     self.seqmap.bases_findMatch = configuration['bases_findMatch']
    #     self.seqmap.histogram_matches(export=True)

    # def sequencing_mapping_to_files(self, minimal_number_of_matching_points):
    #     self.seqmap.give_matches_to_files(match_threshold=minimal_number_of_matching_points)

    # def map_sequencing_and_stage_coordinates(self):
    #     tiles = self.sequencing_data.tiles
    #     self.stage_to_sequencing_mappings = []
    #
    #     for tile in tiles:
    #         matched_files_in_tile = [file for file in self.files if file.sequencing_match and
    #                                  file.sequencing_match.tile == tile.number]
    #
    #         if len(matched_files_in_tile) == 0:
    #             self.stage_to_sequencing_mappings.append(None)
    #             continue
    #
    #         stage_coordinates = np.array([file.movie.stage_coordinates[0] for file in matched_files_in_tile])
    #         # stage_coordinates_in_pixels = np.array([file.movie.stage_coordinates_in_pixels[0] for file in matched_files_in_tile])
    #         sequencing_coordinates = np.vstack(
    #             [file.sequencing_match.transform_coordinates(np.array([[0, 0]])) for file in matched_files_in_tile])
    #
    #         stage_to_tile_mapping = \
    #             mp.MatchPoint(stage_coordinates, sequencing_coordinates, transformation_type='linear',
    #                      source_name='Stage coordinates', destination_name='Sequencing_coordinates',
    #                      source_unit='μm', destination_unit='FASTQ',
    #                      name=f'Stage to sequencing coordinates - tile {tile.number}')
    #         stage_to_tile_mapping.direct_match('linear')
    #         stage_to_tile_mapping.tile = tile.number
    #
    #         stage_to_tile_mapping.save(self.main_path.joinpath(f'stage_to_tile_{tile.number}.mapping'))
    #         self.stage_to_sequencing_mappings.append(stage_to_tile_mapping)
    #
    # def show_stage_to_sequencing_mappings(self):
    #     for mapping in self.stage_to_sequencing_mappings:
    #         if mapping is not None:
    #             mapping.show_mapping_transformation(source_colour='forestgreen', destination_colour='k',
    #                                                 save_path=self.main_path)

    # def tile_boundaries_in_stage_coordinates(self):
    #     # TODO: Use real tile boundaries here
    #     boundaries = []
    #     for mapping in self.stage_to_sequencing_mappings:
    #         if mapping:
    #             boundaries.append(mapping.transform_coordinates([[0, 0], [30000, 30000]], inverse=True).T)
    #         else:
    #             boundaries.append(np.zeros((2,2)))
    #     return np.sort(boundaries)

    def files_with_sequencing_match(self, files=None):
        if files is None:
            files = self.files
        from objectlist import ObjectList
        return ObjectList([file for file in files
                           if file.absoluteFilePath.with_name(file.name + '_sequencing_match.nc').is_file()])

    def sequencing_matches(self, files=None):
        if files is None:
            files = self.files
        from .mapping_collection import MappingCollection
        return MappingCollection([file.sequencing_match for file in files if file.sequencing_match is not None])

    # TODO: Check this method
    def show_sequencing_matches(self, show_file_coordinates=False):
        plot_matched_files_in_tile(self.files_with_sequencing_match, show_file_coordinates=show_file_coordinates,
                                   save=True)

    # TODO: Check this method
    def sequencing_match_info_per_file(self):
        columns = pd.MultiIndex.from_product([['File coordinates', 'Sequencing coordinates'],
                                              ['Matched', 'Total', 'Fraction']])
        df = pd.DataFrame(columns=columns)

        for file in self.files_with_sequencing_match():
            df.loc[file.name, ('File coordinates', 'Total')] = file.sequencing_match.source_cropped.shape[0]
            df.loc[file.name, ('Sequencing coordinates', 'Total')] = file.sequencing_match.destination_cropped.shape[0]
            df.loc[file.name, (['File coordinates','Sequencing coordinates'], 'Matched')] = \
                file.sequencing_match.number_of_matched_points
            df.loc[file.name, ('File coordinates', 'Fraction')] = \
                df.loc[file.name, ('File coordinates', 'Matched')] / \
                df.loc[file.name, ('File coordinates', 'Total')]
            df.loc[file.name, ('Sequencing coordinates', 'Fraction')] = \
                df.loc[file.name, ('Sequencing coordinates', 'Matched')] / \
                df.loc[file.name, ('Sequencing coordinates', 'Total')]

        return df

    # TODO: Check this method
    def sequencing_match_info_mean(self, distance_threshold=25):
        df = self.sequencing_match_info_per_file(distance_threshold)
        df = pd.DataFrame([df.mean(axis=0), df.std(axis=0)], index=['Mean', 'Std']).T
        return df.apply(std_string, axis=1)

    # # TODO: Put this improvement in file
    # def sequencing_mapping_improvement(self):
    #     for match in self.seqmap.matches:
    #         match.nearest_neighbour_match(distance_threshold=25)


# def map_sequences_to_molecules(files, sequencing_data_for_mapping, mapping_sequence, tile, write_path, match_threshold = 5):
#
#

# tile = sequencing_data_for_mapping.get_tile_object(tile=tile)
#
# seqmap = SequencingDataMapping(tile, files, 'similarity',
#                                write_path, nBins=250,
#                                rotationRange=np.array([-180, 180]) / 360 * 2 * np.pi,
#                                magnificationRange=np.array([3000, 6000]))
# #                               rotationRange = np.array([-5,5])/360*2*np.pi+np.pi/2,
# #                               magnificationRange = np.array([2750,3000]))
# # seqmap.initial_image_transformation = {'reflection': 0, 'rotation': -35/180*np.pi, 'magnification':1/125 }
# # seqmap.initial_image_transformation = {'reflection': 0, 'rotation': np.pi, 'magnification': 2*3.336}
# seqmap.initial_image_transformation = {'reflection': 0, 'rotation': np.pi, 'magnification': 3.336}
# # seqmap.initial_image_transformation = {}
# seqmap.bases_findMatch = 500
# seqmap.histogram_matches(export=True)
# seqmap.give_matches_to_files()

# matches = [match for match in seqmap.matches if match.count >= match_threshold]
# # matches = [match for match in seqmap.matches if match.percentMatch > 0.69]
# for match in matches:
#     match.nearest_neighbour_match(distance_threshold=25)
#     seqmap.plot_match(match)

# return seqmap, matches

# Probably it would be better to move the function below somewhere else [IS 12-02-2020]
def within_bounds(coordinates, bounds, margin=0):
    bounds = np.sort(bounds)
    criteria = np.array([(coordinates[:, 0] > (bounds[0, 0] + margin)),
                         (coordinates[:, 0] < (bounds[1, 0] - margin)),
                         (coordinates[:, 1] > (bounds[0, 1] + margin)),
                         (coordinates[:, 1] < (bounds[1, 1] - margin))
                         ])

    return criteria.all(axis=0)


class File:
    # def map_file_sequences_to_molecules(self, mapping_sequence, tile, match_threshold=5):
    #     sequencing_mapping_path = self.experiment.main_path.joinpath('sequencing_mapping')
    #     sequencing_mapping_path.mkdir(exist_ok=True)
    #     # Probably we will not have to export seqmap, at least not to file.
    #     self.seqmap, self.matches = map_sequences_to_molecules([self], self.experiment.sequencing_data, mapping_sequence,
    #                                                             tile, sequencing_mapping_path, match_threshold)
    #

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._sequencing_data = None
        self._sequencing_match = None
        # self.sequencing_match_old = None

        self.dataset_variables += ['sequence', 'sequence_coordinates', 'sequence_quality', 'sequence_tile', 'sequence_variable']

        # if self.experiment.import_all is True:
        #     self.findAndAddExtensions()

    def __getstate__(self):
        self._sequencing_data = None
        return super().__getstate__()

    @property
    @return_none_when_executed_by_pycharm
    def sequencing_match(self):
        if self._sequencing_match is None:
            self.import_sequencing_match()

        return self._sequencing_match

    @sequencing_match.setter
    def sequencing_match(self, value):
        self._sequencing_match = value
        self.export_sequencing_match()

    @property
    @return_none_when_executed_by_pycharm
    def has_sequencing_match(self):
        filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_match')
        # TODO: Link this to the possible file formats of MatchPoint
        for suffix in ['.nc', '.mapping']:
            filepath = filepath.with_suffix(suffix)
            if filepath.is_file():
                return True
        else:
            return False


    @property
    @return_none_when_executed_by_pycharm
    def sequencing_data(self):
        filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_data.nc')
        if filepath.is_file():
            return SequencingData(filepath)
        else:
            return None

    @sequencing_data.setter
    def sequencing_data(self, value):
        filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_data.nc')
        if value is None:
            filepath.unlink(missing_ok=True)
        else:
            value.save(filepath)

    def get_sequencing_data(self, margin=1, mapping_name='All files'):
        # TODO: Make it faster, perhaps by first checking in which tile the file is located.
        sequencing_dataset = self.experiment.sequencing_data

        data_var_names = sequencing_dataset.dataset.data_vars.keys()
        if 'x_sm' not in data_var_names or 'y_sm' not in data_var_names:
            self.experiment.transform_sequencing_to_single_molecule_coordinates()

        x_lims, y_lims = self.movie.boundaries_stage.T

        x_sm = sequencing_dataset.x_sm.sel(mapping_name=mapping_name)
        y_sm = sequencing_dataset.y_sm.sel(mapping_name=mapping_name)

        selection = (x_sm > (x_lims[0] - margin)) & (x_sm < (x_lims[1] + margin)) & \
                    (y_sm > (y_lims[0] - margin)) & (y_sm < (y_lims[1] + margin))\
                        .reset_coords('mapping_name', drop=True)

        if selection.any():
            sequencing_data = sequencing_dataset[selection.values]
            sequencing_data.dataset = sequencing_data.dataset.sel(mapping_name=mapping_name)
            sequencing_data.dataset = sequencing_data.dataset.assign_coords(sequence_in_file=('sequence', np.arange(len(sequencing_data))))
            sequencing_data.dataset.load()
            self.sequencing_data = sequencing_data
        else:
            self.sequencing_data = None

    def generate_sequencing_match(self, overlapping_points_threshold=25, excluded_sequence_names=None, plot=False):
        sequencing_data = self.sequencing_data # Load sequencing data

        if sequencing_data is None:
            # raise AttributeError('Sequencing data not defined, run get_sequencing_data() first')
            self.sequencing_match = None
            return

        if excluded_sequence_names is not None:
            selection = ~np.any(np.vstack([sequencing_data.reference_name == name for name in excluded_sequence_names]), axis=0)
        else:
            selection = np.ones(len(sequencing_data)).astype(bool)

        sequencing_data.dataset['selected'] = xr.DataArray(selection, coords={'sequence': sequencing_data.dataset.sequence})
        self.sequencing_data = sequencing_data

        sequencing_data = sequencing_data[selection]

        source = self.coordinates_stage
        destination = sequencing_data.dataset[['x_sm','y_sm']].to_array('dimension').T

        if source.shape[0] < overlapping_points_threshold or destination.shape[0] < overlapping_points_threshold:
            self.sequencing_match = None
            return

        mapping = mp.MatchPoint(source, destination, transformation_type='similarity')
        mapping.transformation = SimilarityTransform()
        mapping.source_name = 'Single-molecule coordinates'
        mapping.destination_name = 'Sequencing coordinates'
        mapping.source_unit = mapping.destination_unit = 'µm'

        if mapping.source_cropped.shape[0] > overlapping_points_threshold and \
                mapping.destination_cropped.shape[0] > overlapping_points_threshold:
            self.sequencing_match = mapping
            if plot:
                mapping.show_mapping_transformation()
        else:
            self.sequencing_match = None

    def insert_sequencing_data_into_file_dataset(self, include_raw_sequences=False, include_aligned_sequences=True,
                                                 include_sequence_subset=True, determine_matched_pairs=True,
                                                 include_aligned_position=False):
        #TODO: Extend the dataset independent of whether sequencing data is present or not???
        sequencing_data = self.sequencing_data # Load sequencing data
        if self.sequencing_match is None:
            #selected_sequencing_data = None
            return None
        else:
            if self.sequencing_match.destination_distance_threshold == 0:
                raise RuntimeError('No distance threshold set in sequencing match for pair determination')

            if determine_matched_pairs:
                self.sequencing_match.determine_matched_pairs()
            single_molecule_indices, sequence_indices = self.sequencing_match.matched_pairs.T

            selected_sequencing_data = sequencing_data[sequencing_data.selected][sequence_indices]

        sequencing_dataset = xr.Dataset(coords={'molecule_in_file': ('molecule', self.molecule_in_file.values)})

        sequence_name_dtype = sequencing_data.dataset.reference_name.astype('S').dtype
        variable_dict = {
            'sequence_name': ('reference_name', np.array('').astype(sequence_name_dtype)),
            'sequence_tile': ('tile', np.int64(0)),
            'sequence_in_file': ('sequence_in_file', -1) # -1 is not ideal as it will not give an error when used as index, but currently I don't have another solution
        }

        sequence_length = len(self.sequencing_data.read1_sequence[0].item())

        if 'sample_name' in sequencing_data.dataset.data_vars:
            sequence_sample_name_dtype = sequencing_data.dataset.sample_name.astype('S').dtype
            variable_dict['sequence_sample_name'] = ('sample_name', np.array('').astype(sequence_sample_name_dtype))

        if include_raw_sequences:
            variable_dict['sequence'] = ('read1_sequence', b'-' * sequence_length)
            variable_dict['sequence_quality'] = ('read1_quality', b' ' * sequence_length)

        if include_aligned_sequences and hasattr(self.sequencing_data, 'read1_sequence_aligned'):
            variable_dict['sequence_aligned'] = ('read1_sequence_aligned', b'-' * sequence_length)
            variable_dict['sequence_quality_aligned'] = ('read1_quality_aligned', b' ' * sequence_length)

        if include_sequence_subset and hasattr(self.sequencing_data, 'sequence_subset'):
            subset_length = len(self.sequencing_data.sequence_subset[0].item())
            variable_dict['sequence_subset'] = ('sequence_subset', b'-' * subset_length)
            variable_dict['sequence_quality_subset'] = ('quality_subset', b' ' * subset_length)

        if include_aligned_position and hasattr(self.sequencing_data, 'position'):
            variable_dict['sequence_aligned_position'] = ('position', -1)

        is_empty = selected_sequencing_data is None or (len(selected_sequencing_data) == 0)

        for key_file, (key_sequencing_data, default_value) in variable_dict.items():
            sequencing_dataset[key_file] = ('molecule', np.full(self.number_of_molecules, default_value))
            if not is_empty:
                sequencing_dataset[key_file][single_molecule_indices] = selected_sequencing_data.dataset[key_sequencing_data]

        sequencing_dataset = sequencing_dataset.set_coords('sequence_in_file')

        sequencing_dataset['sequence_coordinates'] = (('molecule', 'dimension'), np.zeros((self.number_of_molecules, 2), 'int64'))
        if not is_empty:
            sequencing_dataset['sequence_coordinates'][single_molecule_indices] = selected_sequencing_data.coordinates
        # sequencing_dataset['dimension'] = [b'x', b'y']

        sequencing_dataset.to_netcdf(self.relativeFilePath.with_suffix('.nc'), engine='netcdf4', mode='a')

    # @property
    # def sequences(self):
    #     if len(self.molecules) > 0:
    #         return np.vstack([molecule.sequence for molecule in self.molecules])
    #     else:
    #         return np.array([])

    # @sequences.setter
    # def sequences(self, sequences):
    #     for i, molecule in enumerate(self.molecules):
    #         molecule.sequence = sequences[i]

    # @property
    # def sequence_indices(self):
    #     if len(self.molecules) > 0:
    #         return np.array([molecule.sequence_index for molecule in self.molecules])
    #     else:
    #         return np.array([])

    # @sequence_indices.setter
    # def sequence_indices(self, sequence_indices):
    #     for i, molecule in enumerate(self.molecules):
    #         molecule.sequence_index = sequence_indices[i]

    # @property
    # def sequencing_tile(self):
    #     if self.sequencing_data:
    #         if len(self.sequencing_data.tiles)==1:
    #             return self.sequencing_data.tiles[0]
    #         else:
    #             raise ValueError('File contains sequences from multiple tiles')
    #     else:
    #         return None

    # @property
    # def sequencing_data(self):
    #     return self._sequencing_data
    #
    # @sequencing_data.setter
    # def sequencing_data(self, sequencing_data):
    #     # self.molecules = []
    #     # self.number_of_molecules = len(sequencing_data)
    #     # for i, molecule in enumerate(self.molecules):
    #     #     molecule.sequencing_data = sequencing_data[i]

    # def import_sequencing_data(self):
    #     self.sequencing_data = FastqData(self.absoluteFilePath.with_suffix('.fastq'))
    #     # self.sequences = self.sequencing_data.sequence

    def import_sequencing_match(self):
        filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_match')
        # TODO: Link this to the possible file formats of MatchPoint
        for suffix in ['.nc', '.mapping']:
            filepath = filepath.with_suffix(suffix)
            if filepath.is_file():
                self._sequencing_match = mp.MatchPoint.load(filepath)
                break
        else:
            self._sequencing_match = None

    def export_sequencing_match(self):
        filepath = self.absoluteFilePath.with_name(self.name+'_sequencing_match.nc')
        if self._sequencing_match is None:
            # TODO: Link this to the possible file formats of MatchPoint
            for suffix in ['.nc', '.mapping']:
                filepath = filepath.with_suffix(suffix)
                filepath.unlink(missing_ok=True)
        else:
            self._sequencing_match.save(filepath)

    # def import_sequencing_data(self):
    #     filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_data.nc')
    #     if filepath.is_file():
    #         self._sequencing_data = SequencingData(filepath)
    #     else:
    #         self._sequencing_data = None
    #
    #
    # def export_sequencing_data(self):
    #     filepath = self.absoluteFilePath.with_name(self.name + '_sequencing_data.nc')
    #     if self._sequencing_data is None:
    #         filepath.unlink(missing_ok=True)
    #     else:
    #         self._sequencing_data.save(filepath)

        # if self.sequencing_match_old is not None:
        #     self.sequencing_match_old.save(self.absoluteFilePath.with_name(self.name + '_sequencing_match_old.mapping'))

    # def find_sequences(self, maximum_distance_file, tuple_size, initial_transformation={},
    #                    hash_table_distance_threshold=0.01,
    #                    alpha=0.1, test_radius=10, K_threshold=10e9,
    #                    magnification_range=None, rotation_range=None,
    #                    channel=0,
    #                    nearest_neighbour_match_distance_threshold=25):
    #     # TODO: Make geometric hashing reflection invariant
    #     initial_transformation = AffineTransform(**initial_transformation)
    #
    #     # TODO: make the following line more general and remove bounds dependence in geometric hashing
    #     source_vertices = self.movie.channel_vertices(channel)
    #     coordinate_vertices_file = initial_transformation(source_vertices)
    #
    #     #self.geometric_hash_data = geometric_hash(initial_transform(self.coordinates), maximum_distance_file, tuple_size)
    #
    #     coordinates = self.coordinates_from_channel(channel)
    #
    #     #match.destination_index = destination_index
    #     match = find_match_after_hashing(initial_transformation(coordinates), maximum_distance_file, tuple_size, coordinate_vertices_file,
    #                                      *self.experiment.geometric_hash_data,
    #                                      hash_table_distance_threshold, alpha, test_radius, K_threshold,
    #                                      magnification_range, rotation_range)
    #     if match:
    #         match.tile = self.experiment.sequencing_data_for_mapping.tiles[match.destination_index].number
    #         match.channel = channel
    #         match.source = self.coordinates
    #         match.initial_transformation = initial_transformation
    #         match.transformation = match.transformation @ initial_transformation.params
    #         match.source_vertices = source_vertices
    #         match.calculate_inverse_transformation()
    #         # TODO: Base this on some better criteria
    #         #match.nearest_neighbour_match(nearest_neighbour_match_distance_threshold)
    #         self.sequencing_match = match
    #         self.export_sequencing_match()
    #         #self.get_all_sequences_from_sequencing_data()
    #
    # def find_sequences3(self, distance=15, alpha=0.9, sigma=10, K_threshold=10e2, channel=0,
    #                     nearest_neighbour_match_distance_threshold=25):
    #
    #     coordinates = self.coordinates_from_channel(channel)
    #
    #     match = self.experiment.geometric_hashtable.query(coordinates, distance, alpha, sigma, K_threshold)
    #
    #     if match:
    #         # match.destination_index = 0
    #         match.tile = self.experiment.sequencing_data_for_mapping.tiles[match.destination_index].number
    #         match.channel = channel
    #
    #         # TODO: Base this on some better criteria
    #         #match.nearest_neighbour_match(nearest_neighbour_match_distance_threshold)
    #         self.sequencing_match = match
    #         self.export_sequencing_match()
    #         #self.get_all_sequences_from_sequencing_data()

    def find_sequences_using_stage_coordinates(self, channel=0, show=False, save=True):
        boundaries = self.experiment.tile_boundaries_in_stage_coordinates()
        tile_selection = np.array([self.movie.stage_coordinates > boundaries[:, :, 0],
                                   self.movie.stage_coordinates < boundaries[:, :, 1]])
        tile_index = np.where(np.all(tile_selection, axis=(0, 2)))[0][0]

        source_vertices = self.movie.channels[channel].vertices
        source_vertices_in_stage_coordinates = \
            self.movie.stage_coordinates + np.flip(source_vertices * self.movie.pixel_size, axis=1)


        source_vertices_in_sequencing_coordinates = \
            self.experiment.stage_to_sequencing_mappings[tile_index]\
                .transform_coordinates(source_vertices_in_stage_coordinates)

        match = mp.MatchPoint(source_vertices, source_vertices_in_sequencing_coordinates, transformation_type='linear')
        match.direct_match()
        match.source_vertices = source_vertices
        match.destination_vertices = np.array([[1720, 1330], [29720, 1330], [29720, 29330], [1720, 29330]]) # Tile coordinate bounds MiSeq
        match.tile = self.experiment.sequencing_data_for_mapping.tiles[tile_index].number
        match.channel = channel

        match.source = self.coordinates_from_channel(channel)
        match.destination = self.experiment.sequencing_data_for_mapping.tiles[tile_index].coordinates

        self.sequencing_match = match
        if show:
            self.plot_sequencing_match()
        if save:
            self.export_sequencing_match()


    # def get_sequencing_data_for_file(self): #, margin=10):
    #
    #     # sequencing_coordinates_in_image = \
    #     #     self.sequencing_match.transform_coordinates(self.experiment.sequencing_data.coordinates, inverse=True)
    #     # coordinate_selection = \
    #     #     pth.Path(self.sequencing_match.source_vertices).contains_points(sequencing_coordinates_in_image)
    #     #
    #     # self.sequencing_data = \
    #     #     self.experiment.sequencing_data.get_selection(tile=self.sequencing_match.tile,
    #     #                                                   boolean_selection=coordinate_selection)
    #
    #     def interpolate(coordinates):
    #         coordinates = np.vstack([coordinates, coordinates[0]])
    #         return np.linspace(coordinates, np.roll(coordinates, -1, axis=0)).reshape(-1, 2, order='F')
    #
    #     source_vertices_interpolated = interpolate(self.sequencing_match.source_vertices)
    #     destination_vertices_interpolated = self.sequencing_match.transform_coordinates(source_vertices_interpolated)
    #     self.sequencing_data = \
    #         self.experiment.sequencing_data.get_selection(tile=self.sequencing_match.tile,
    #                                                       coordinates_within_vertices=destination_vertices_interpolated)
    #
    #     self.sequencing_data.export_fastq(self.relativeFilePath)

    # def optimize_sequencing_match_using_visible_sequences(self, visible_sequence_names='', distance_threshold=25,
    #                                                       show=False, save=True, **kwargs):
    #     self.sequencing_match_old = copy.deepcopy(self.sequencing_match)
    #
    #     sequencing_data = self.sequencing_data.get_selection(in_name=visible_sequence_names)
    #
    #     self.sequencing_match.source = self.coordinates_from_channel(self.sequencing_match.channel)
    #     self.sequencing_match.destination = sequencing_data.coordinates
    #     self.sequencing_match.nearest_neighbour_match(distance_threshold, **kwargs)
    #
    #     if show:
    #         self.plot_sequencing_match()
    #     if save:
    #         self.export_sequencing_match()

    def determine_sequences_at_current_coordinates(self, visible_sequence_names='', distance_threshold=None):
        selection = self.sequencing_data.selection(in_name=visible_sequence_names)

        sequence_coordinates_in_file = \
            self.sequencing_match.transform_coordinates(self.sequencing_data.coordinates[selection], inverse=True)

        distances, molecule_indices, sequence_indices = \
            mp.icp.nearest_neighbor_pair(self.coordinates_from_channel(self.sequencing_match.channel),
                                  sequence_coordinates_in_file,
                                  distance_threshold=distance_threshold)

        cum_selection = np.cumsum(selection)-1

        # for molecule_index, sequence_index, distance in zip(molecule_indices, sequence_indices, distances):
        #     self.molecules[molecule_index].sequence_index = np.argmax(cum_selection == sequence_index)
        #     self.molecules[molecule_index].distance_to_sequence = distance

        sequence_indices2 = np.array([np.argmax(cum_selection == sequence_index)
                                     for molecule_index, sequence_index, distance
                                     in zip(molecule_indices, sequence_indices, distances)])

        #molecule_index = self.molecule.to_index()[molecule_indices]
        selected_sequencing_data = self.sequencing_data[sequence_indices2]
        molecule = self.molecule_in_file
        sequencing_dataset = xr.Dataset(
            {
                'sequence':             (('molecule', 'nucleotide'), selected_sequencing_data.sequence),
                'quality':              (('molecule', 'nucleotide'), selected_sequencing_data.quality),
                'distance_to_sequence': ('molecule', distances)
            },
            coords=
            {
                'sequence_name':    ('molecule', selected_sequencing_data.name),
                'molecule':         ('molecule', molecule_indices),
                'sequence_in_file': ('molecule', sequence_indices2)
            }
        )

        sequencing_dataset = sequencing_dataset.reindex_like(molecule.set_index(molecule='molecule_in_file'),
                                        fill_value={'sequence_name': '',
                                                    'sequence_in_file': np.array(np.nan).astype(pd.UInt16Dtype)})
        sequencing_dataset = sequencing_dataset.reset_index('molecule').rename(molecule_='molecule_in_file')

        # Engine netcdf4 has some locking problems.
        sequencing_dataset.to_netcdf(self.relativeFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')

    def use_sequences_as_molecules(self):
        self.molecules = []
        file_coordinates = self.sequencing_match.transform_coordinates(self.sequencing_data.coordinates, inverse=True)
        self.set_coordinates_of_channel(file_coordinates, channel=self.sequencing_match.channel)

    def plot_sequencing_match(self):
        #plot_sequencing_match(self.sequencing_match)
        #name = f'Tile: {self.sequencing_tile.name}, File: {str(self.relativeFilePath)}'
        filename = self.name + '_sequencing_mapping'
        print(filename)

        title = f'File: {self.relativePath.joinpath(self.name)} - Tile: {self.sequencing_match.tile}'

        # TODO: Put this somewhere, where it makes sense.
        def MiSeq_pixels_to_um(pixels):
            return 958 / 2800 * (pixels - 1000) / 10

        # TODO: Put this somewhere, where it makes sense.
        # TODO: Make this more general, get this value from tif metadata
        def Fluo_pixels_to_um(pixels):
            return pixels * 0.125

        plot_sequencing_match(self.sequencing_match, self.absoluteFilePath.parent, title, filename,
                              'um', MiSeq_pixels_to_um, Fluo_pixels_to_um)


    # mp.icp.show_point_connections(self.coordinates[molecule_indices], sequence_coordinates_in_file[sequence_indices])


    # This is probably not the way to go
    # def give_molecules_closest_sequence(self):
    #     sequencing_data = self.experiment.sequencing_data
    #     indices_within_tile = sequencing_data.selection(tile=int(self.sequence_match.tile.name))
    #     sequencing_data = sequencing_data.select(indices_within_tile, copyData=True)
    #
    #     #raise Warning("Not implemented for the donor channel yet")
    #     boundaries = self.sequence_match.transform_coordinates(self.movie.channel_boundaries('a'))
    #
    #     indices_within_bounds = within_bounds(sequencing_data.coordinates, boundaries)
    #     sequencing_data.select(indices_within_bounds, copyData=False)
    #
    #     tile_coordinates = sequencing_data.coordinates
    #     image_coordinates = self.sequence_match.transform_coordinates(self.coordinates)
    #
    #     from papylio.plotting import scatter_coordinates
    #     plt.figure()
    #     scatter_coordinates([tile_coordinates, image_coordinates])
    #
    #
    #     distances, image_indices, tile_indices = mp.icp.nearest_neighbor_pair(image_coordinates, tile_coordinates)
    #
    #     distance_threshold = 25
    #     image_indices = image_indices[distances < distance_threshold]
    #     tile_indices = tile_indices[distances < distance_threshold]
    #
    #     print(*[sequence.tostring() for sequence in sequencing_data.sequence[tile_indices]], sep="\n")
    #
    #     plt.figure()
    #     scatter_coordinates([tile_coordinates, self.sequence_match.destination, image_coordinates])
    #     mp.icp.show_point_connections(image_coordinates[image_indices], tile_coordinates[tile_indices])
    #
    #     for molecule in self.molecules: molecule.sequence = np.array([], dtype=bytes)
    #     for i in np.arange(len(image_indices)):
    #         self.molecules[image_indices[i]].sequence = sequencing_data.sequence[tile_indices[i]]
    #         self.molecules[image_indices[i]].sequencing_data = sequencing_data.select(i, copyData=True)


class Molecule:
    slots = 'sequence_index'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sequence_index = None

    @property
    @return_none_when_executed_by_pycharm
    def sequence_name(self):
        if self.sequence_index is not None:
            return self.file.sequencing_data.name[self.sequence_index]

    @property
    @return_none_when_executed_by_pycharm
    def sequence(self):
        if self.sequence_index is not None:
            return self.file.sequencing_data.sequence[self.sequence_index, :].tobytes().decode('UTF-8')

    @property
    @return_none_when_executed_by_pycharm
    def sequencing_data(self):
        if self.sequence_index is not None:
            return self.file.sequencing_data[self.sequence_index]


def std_string(value_and_uncertainty):
    value, uncertainty = value_and_uncertainty
    exponent = math.floor(math.log10(uncertainty))
    decimals = -exponent if exponent < 0 else 0
    units = exponent if exponent > 0 else 1
    return f'{np.round(value,-exponent):{units}.{decimals}f}±{np.round(uncertainty,-exponent):{units}.{decimals}f}'

def mean_and_std_string(values):
    return std_string(np.mean(values), np.std(values))


# -*- coding: utf-8 -*-
"""
Created on Sun Jun  2 01:48:30 2019

@author: iason

Usefull decorators
"""

import functools
import time


# General decorator template
def decorator(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        # Do something before
        value = func(*args, **kwargs)
        # Do something after
        return value
    return wrapper_decorator



def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} sec")
        return value
    return wrapper_timer

@timer
def waste_some_time(num_times):
    for _ in range(num_times):
        sum([i**2 for i in range(10000)])

if __name__ == '__main__':
    waste_some_time(100)


import pytest
from papylio.plugins.sequencing.sequence_generation import generate_sequences
from papylio.plugins.sequencing.sequence_plotting import plot_sequence_density

@pytest.fixture
def sequences():
    return generate_sequences(base_composition=['N', 'AC', 'GC', 'ACTG'])


def test_plot_sequence_density(sequences):
    plot_sequence_density(sequences, expected_seq=None, start=None, end=None, row_length=None,
                          figure=None, save=False, title='')

